from __future__ import annotations

from collections.abc import Mapping
from enum import Enum
from typing import Optional

from rflx import expression as expr
from rflx.ada import (
    TRUE,
    Add,
    Aggregate,
    And,
    AndThen,
    Assignment,
    Call,
    CallStatement,
    ChoiceList,
    Constrained,
    Div,
    Equal,
    Expr,
    ExpressionFunctionDeclaration,
    FormalSubprogramDeclaration,
    FunctionSpecification,
    Greater,
    GreaterEqual,
    If,
    IfStatement,
    In,
    Indexed,
    InlineAlways,
    InOutParameter,
    Length,
    LessEqual,
    Mod,
    Mul,
    NamedAggregate,
    Not,
    Number,
    ObjectDeclaration,
    Old,
    OutParameter,
    Parameter,
    Postcondition,
    Pragma,
    PragmaStatement,
    Precondition,
    ProcedureSpecification,
    Selected,
    Size,
    Slice,
    Statement,
    String,
    Sub,
    SubprogramBody,
    SubprogramDeclaration,
    UnitPart,
    Update,
    ValueRange,
    Variable,
)
from rflx.const import BUILTINS_PACKAGE
from rflx.identifier import ID, StrID
from rflx.model import (
    FINAL,
    ByteOrder,
    Enumeration,
    Field,
    Message,
    Opaque,
    Scalar,
    Sequence,
    Type,
    is_builtin_type,
)

from . import common, const


class SerializerGenerator:
    def __init__(self, prefix: str = "") -> None:
        self.prefix = prefix

    def create_valid_size_function(self, message: Message) -> UnitPart:
        specification = FunctionSpecification(
            "Valid_Size",
            "Boolean",
            [
                Parameter(["Ctx"], "Context"),
                Parameter(["Fld"], "Field"),
                Parameter(["Size"], const.TYPES_BIT_LENGTH),
            ],
        )

        implicit_size_condition = LessEqual(
            Variable("Size"),
            Call(
                "Available_Space",
                [Variable("Ctx"), Variable("Fld")],
            ),
        )
        explicit_size_condition = Equal(
            Variable("Size"),
            Call("Field_Size", [Variable("Ctx"), Variable("Fld")]),
        )

        return UnitPart(
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    (
                        If(
                            [
                                (
                                    AndThen(
                                        Equal(Variable("Fld"), Variable(l.target.affixed_name)),
                                        *(
                                            [
                                                l.condition.substituted(
                                                    common.substitution(
                                                        message,
                                                        self.prefix,
                                                        target_type=const.TYPES_BIT_LENGTH,
                                                    ),
                                                )
                                                .simplified()
                                                .ada_expr(),
                                            ]
                                            if l.condition != expr.TRUE
                                            else []
                                        ),
                                    ),
                                    implicit_size_condition,
                                )
                                for l in message.structure
                                if l.has_implicit_size
                            ],
                            explicit_size_condition,
                        )
                        if len(message.fields) > 1
                        or all(not l.has_implicit_size for l in message.structure)
                        else implicit_size_condition
                    ),
                    [
                        Precondition(
                            And(
                                Call(
                                    self.prefix * message.identifier * "Valid_Next",
                                    [Variable("Ctx"), Variable("Fld")],
                                ),
                            ),
                        ),
                    ],
                ),
            ],
        )

    def create_valid_length_function(self, message: Message) -> UnitPart:
        specification = FunctionSpecification(
            "Valid_Length",
            "Boolean",
            [
                Parameter(["Ctx"], "Context"),
                Parameter(["Fld"], "Field"),
                Parameter(["Length"], const.TYPES_LENGTH),
            ],
        )

        return UnitPart(
            [
                # Eng/RecordFlux/Workarounds#47
                Pragma(
                    "Warnings",
                    [Variable("Off"), String("postcondition does not mention function result")],
                ),
                SubprogramDeclaration(
                    specification,
                    [
                        Precondition(
                            And(
                                Call(
                                    self.prefix * message.identifier * "Valid_Next",
                                    [Variable("Ctx"), Variable("Fld")],
                                ),
                            ),
                        ),
                        Postcondition(TRUE),
                    ],
                ),
                Pragma(
                    "Warnings",
                    [Variable("On"), String("postcondition does not mention function result")],
                ),
            ],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    Call(
                        "Valid_Size",
                        [
                            Variable("Ctx"),
                            Variable("Fld"),
                            Call(
                                const.TYPES_TO_BIT_LENGTH,
                                [Variable("Length")],
                            ),
                        ],
                    ),
                ),
            ],
        )

    @staticmethod
    def requires_set_procedure(message: Message) -> bool:
        return any(
            message.is_possibly_empty(f) or isinstance(t, (Scalar, Sequence))
            for f, t in message.field_types.items()
        )

    def create_set_procedure(
        self,
        message: Message,
        composite_fields: list[Field],
    ) -> UnitPart:
        if not self.requires_set_procedure(message):
            return UnitPart()

        def specification() -> ProcedureSpecification:
            return ProcedureSpecification(
                "Set",
                [
                    InOutParameter(["Ctx"], "Context"),
                    Parameter(["Fld"], "Field"),
                    Parameter(["Val"], const.TYPES_BASE_INT),
                    Parameter(["Size"], const.TYPES_BIT_LENGTH),
                    Parameter(["State_Valid"], "Boolean"),
                    OutParameter(["Buffer_First"], const.TYPES_INDEX),
                    OutParameter(["Buffer_Last"], const.TYPES_INDEX),
                    OutParameter(["Offset"], const.TYPES_OFFSET),
                ],
            )

        class CursorState(Enum):
            VALID = 1
            WELL_FORMED = 2

        def set_context_cursor(field_type: CursorState) -> Assignment:
            return Assignment(
                Indexed(Variable("Ctx.Cursors"), Variable("Fld")),
                NamedAggregate(
                    (
                        "State",
                        (
                            Variable("S_Valid")
                            if field_type == CursorState.VALID
                            else Variable("S_Well_Formed")
                        ),
                    ),
                    ("First", Variable("First")),
                    ("Last", Variable("Last")),
                    ("Value", Variable("Val")),
                ),
            )

        return UnitPart(
            body=[
                SubprogramBody(
                    specification(),
                    [
                        ObjectDeclaration(["First"], const.TYPES_BIT_INDEX),
                        ObjectDeclaration(["Last"], const.TYPES_BIT_LENGTH),
                    ],
                    [
                        CallStatement(
                            "Reset_Dependent_Fields",
                            [Variable("Ctx"), Variable("Fld")],
                        ),
                        Assignment(
                            "First",
                            Call("Field_First", [Variable("Ctx"), Variable("Fld")]),
                        ),
                        Assignment(
                            "Last",
                            Add(
                                Call(
                                    "Field_First",
                                    [Variable("Ctx"), Variable("Fld")],
                                ),
                                Variable("Size"),
                                -Number(1),
                            ),
                        ),
                        Assignment(
                            "Offset",
                            Call(
                                const.TYPES_OFFSET,
                                [
                                    Mod(
                                        Sub(
                                            Size(const.TYPES_BYTE),
                                            Mod(Variable("Last"), Size(const.TYPES_BYTE)),
                                        ),
                                        Size(const.TYPES_BYTE),
                                    ),
                                ],
                            ),
                        ),
                        Assignment(
                            "Buffer_First",
                            Call(const.TYPES_TO_INDEX, [Variable("First")]),
                        ),
                        Assignment(
                            "Buffer_Last",
                            Call(const.TYPES_TO_INDEX, [Variable("Last")]),
                        ),
                        *self._update_last(),
                        IfStatement(
                            [
                                (
                                    Variable("State_Valid"),
                                    [set_context_cursor(CursorState.VALID)],
                                ),
                            ],
                            [set_context_cursor(CursorState.WELL_FORMED)],
                        ),
                        PragmaStatement(
                            "Assert",
                            [
                                Equal(
                                    Variable("Last"),
                                    Sub(
                                        Add(
                                            Call("Field_First", [Variable("Ctx"), Variable("Fld")]),
                                            Variable("Size"),
                                        ),
                                        Number(1),
                                    ),
                                ),
                            ],
                        ),
                    ],
                    [
                        Precondition(
                            AndThen(
                                Call(
                                    self.prefix * message.identifier * "Has_Buffer",
                                    [Variable("Ctx")],
                                ),
                                Call(
                                    self.prefix * message.identifier * "Valid_Next",
                                    [Variable("Ctx"), Variable("Fld")],
                                ),
                                Call(
                                    self.prefix * message.identifier * "Valid_Value",
                                    [Variable("Fld"), Variable("Val")],
                                ),
                                Call(
                                    self.prefix * message.identifier * "Valid_Size",
                                    [Variable("Ctx"), Variable("Fld"), Variable("Size")],
                                ),
                                LessEqual(
                                    Variable("Size"),
                                    Call(
                                        self.prefix * message.identifier * "Available_Space",
                                        [Variable("Ctx"), Variable("Fld")],
                                    ),
                                ),
                                *(
                                    [
                                        If(
                                            [
                                                (
                                                    Call(
                                                        self.prefix
                                                        * message.identifier
                                                        * "Composite_Field",
                                                        [Variable("Fld")],
                                                    ),
                                                    Equal(
                                                        Mod(
                                                            Variable("Size"),
                                                            Size(const.TYPES_BYTE),
                                                        ),
                                                        Number(0),
                                                    ),
                                                ),
                                            ],
                                            Variable("State_Valid"),
                                        ),
                                    ]
                                    if composite_fields
                                    else [
                                        Variable("State_Valid"),
                                    ]
                                ),
                            ),
                        ),
                        Postcondition(
                            AndThen(
                                Call("Valid_Next", [Variable("Ctx"), Variable("Fld")]),
                                *(
                                    [
                                        Call(
                                            "Invalid_Successor",
                                            [Variable("Ctx"), Variable("Fld")],
                                        ),
                                    ]
                                    if len(message.fields) > 1
                                    else []
                                ),
                                Equal(
                                    Variable("Buffer_First"),
                                    Call(
                                        const.TYPES_TO_INDEX,
                                        [
                                            Call(
                                                "Field_First",
                                                [Variable("Ctx"), Variable("Fld")],
                                            ),
                                        ],
                                    ),
                                ),
                                Equal(
                                    Variable("Buffer_Last"),
                                    Call(
                                        const.TYPES_TO_INDEX,
                                        [
                                            Add(
                                                Call(
                                                    "Field_First",
                                                    [Variable("Ctx"), Variable("Fld")],
                                                ),
                                                Variable("Size"),
                                                -Number(1),
                                            ),
                                        ],
                                    ),
                                ),
                                Equal(
                                    Variable("Offset"),
                                    Call(
                                        const.TYPES_OFFSET,
                                        [
                                            Mod(
                                                Sub(
                                                    Size(const.TYPES_BYTE),
                                                    Mod(
                                                        Add(
                                                            Call(
                                                                "Field_First",
                                                                [
                                                                    Variable("Ctx"),
                                                                    Variable("Fld"),
                                                                ],
                                                            ),
                                                            Variable("Size"),
                                                            -Number(1),
                                                        ),
                                                        Size(const.TYPES_BYTE),
                                                    ),
                                                ),
                                                Size(const.TYPES_BYTE),
                                            ),
                                        ],
                                    ),
                                ),
                                *common.context_invariant(message),
                                *[
                                    Equal(e, Old(e))
                                    for e in [
                                        Variable("Ctx.Buffer_First"),
                                        Variable("Ctx.Buffer_Last"),
                                        Variable("Ctx.First"),
                                        Variable("Ctx.Last"),
                                        Call("Has_Buffer", [Variable("Ctx")]),
                                        Call("Field_First", [Variable("Ctx"), Variable("Fld")]),
                                    ]
                                ],
                                Call("Sufficient_Space", [Variable("Ctx"), Variable("Fld")]),
                                If(
                                    [
                                        (
                                            And(
                                                Variable("State_Valid"),
                                                Greater(Variable("Size"), Number(0)),
                                            ),
                                            Call(
                                                "Valid",
                                                [Variable("Ctx"), Variable("Fld")],
                                            ),
                                        ),
                                    ],
                                    Call("Well_Formed", [Variable("Ctx"), Variable("Fld")]),
                                ),
                                self.scalar_setter_and_getter_relation(message),
                                *(
                                    [
                                        common.unchanged_cursor_before_or_invalid(
                                            Variable("Fld"),
                                            loop_entry=False,
                                            or_invalid=False,
                                            including_limit=False,
                                        ),
                                    ]
                                    if len(message.fields) > 1
                                    else []
                                ),
                            ),
                        ),
                    ],
                ),
            ],
        )

    def create_scalar_setter_procedures(
        self,
        message: Message,
        scalar_fields: Mapping[Field, Scalar],
    ) -> UnitPart:
        if not scalar_fields:
            return UnitPart()

        def specification(
            field: Field,
            field_type: Type,
            use_enum_records_directly: bool = False,
        ) -> ProcedureSpecification:
            if field_type.package == BUILTINS_PACKAGE:
                type_identifier = ID(field_type.name)
            elif (
                isinstance(field_type, Enumeration)
                and field_type.always_valid
                and not use_enum_records_directly
            ):
                type_identifier = common.prefixed_type_identifier(
                    common.full_enum_name(field_type),
                    self.prefix,
                )
            else:
                type_identifier = common.prefixed_type_identifier(
                    field_type.identifier,
                    self.prefix,
                )

            return ProcedureSpecification(
                f"Set_{field.name}",
                [InOutParameter(["Ctx"], "Context"), Parameter(["Val"], type_identifier)],
            )

        def precondition(
            field: Field,
            field_type: Scalar,
            use_enum_records_directly: bool = False,
        ) -> Precondition:
            return Precondition(
                AndThen(
                    *self.setter_preconditions(
                        message,
                        self.prefix * message.identifier * field.affixed_name,
                    ),
                    Call(
                        (
                            f"Valid_{field_type.name}"
                            if is_builtin_type(field_type.identifier)
                            else self.prefix * field_type.package * f"Valid_{field_type.name}"
                        ),
                        [
                            (
                                Variable("Val")
                                if use_enum_records_directly
                                else Call(
                                    common.to_base_integer(self.prefix, field_type.package),
                                    [Variable("Val")],
                                )
                            ),
                        ],
                    ),
                    common.sufficient_space_for_field_condition(
                        self.prefix * message.identifier,
                        Variable(self.prefix * message.identifier * field.affixed_name),
                    ),
                    common.field_condition_call(
                        self.prefix,
                        message,
                        field,
                        value=Call(
                            common.to_base_integer(self.prefix, field_type.package),
                            [Variable("Val")],
                        ),
                    ),
                ),
            )

        def postcondition(
            field: Field,
            field_type: Scalar,
            use_enum_records_directly: bool = False,
        ) -> Postcondition:
            return Postcondition(
                And(
                    Call("Has_Buffer", [Variable("Ctx")]),
                    Call(
                        "Valid",
                        [Variable("Ctx"), Variable(field.affixed_name)],
                    ),
                    *(
                        [
                            Equal(
                                Call(f"Get_{field.name}", [Variable("Ctx")]),
                                (
                                    Aggregate(TRUE, Variable("Val"))
                                    if isinstance(field_type, Enumeration)
                                    and field_type.always_valid
                                    and not use_enum_records_directly
                                    else Variable("Val")
                                ),
                            ),
                        ]
                        if int(field_type.value_count) > 1
                        else []
                    ),
                    *self.public_setter_postconditions(message, field),
                    *common.context_cursor_unchanged(message, field, predecessors=True),
                ),
            )

        def body(
            field: Field,
            field_type: Scalar,
            use_enum_records_directly: bool = False,
        ) -> SubprogramBody:
            return SubprogramBody(
                specification(field, field_type, use_enum_records_directly),
                [],
                [
                    CallStatement(
                        "Set_Scalar",
                        [
                            Variable("Ctx"),
                            Variable(field.affixed_name),
                            Call(
                                common.to_base_integer(self.prefix, field_type.package),
                                [Variable("Val")],
                            ),
                        ],
                    ),
                ],
                (
                    [
                        precondition(field, field_type, use_enum_records_directly),
                        postcondition(field, field_type, use_enum_records_directly),
                    ]
                    if use_enum_records_directly
                    else []
                ),
            )

        byte_orders = set(message.byte_order.values())
        uniform_byte_order = byte_orders.pop() if len(byte_orders) == 1 else None

        return UnitPart(
            [
                Pragma(
                    "Warnings",
                    [Variable("Off"), String('aspect "*" not enforced on inlined subprogram "*"')],
                ),
                *[
                    SubprogramDeclaration(
                        specification(f, t),
                        [
                            InlineAlways(),
                            precondition(f, t),
                            postcondition(f, t),
                        ],
                    )
                    for f, t in scalar_fields.items()
                ],
                Pragma(
                    "Warnings",
                    [Variable("On"), String('aspect "*" not enforced on inlined subprogram "*"')],
                ),
            ],
            [
                SubprogramBody(
                    ProcedureSpecification(
                        "Set_Scalar",
                        [
                            InOutParameter(["Ctx"], "Context"),
                            Parameter(["Fld"], "Field"),
                            Parameter(["Val"], const.TYPES_BASE_INT),
                        ],
                    ),
                    [
                        ObjectDeclaration(["Buffer_First", "Buffer_Last"], const.TYPES_INDEX),
                        ObjectDeclaration(["Offset"], const.TYPES_OFFSET),
                        ObjectDeclaration(
                            ["Size"],
                            const.TYPES_BIT_LENGTH,
                            Call("Field_Size", [Variable("Ctx"), Variable("Fld")]),
                            constant=True,
                        ),
                    ],
                    [
                        CallStatement(
                            "Set",
                            [
                                Variable("Ctx"),
                                Variable("Fld"),
                                Variable("Val"),
                                Variable("Size"),
                                TRUE,
                                Variable("Buffer_First"),
                                Variable("Buffer_Last"),
                                Variable("Offset"),
                            ],
                        ),
                        CallStatement(
                            const.TYPES * "Lemma_Size",
                            [
                                Variable("Val"),
                                Call("Positive", [Variable("Size")]),
                            ],
                        ),
                        CallStatement(
                            const.TYPES_OPERATIONS * "Insert",
                            [
                                Variable("Val"),
                                Variable("Ctx.Buffer.all"),
                                Variable("Buffer_First"),
                                Variable("Buffer_Last"),
                                Variable("Offset"),
                                Call("Positive", [Variable("Size")]),
                                (
                                    Variable(
                                        (
                                            const.TYPES_HIGH_ORDER_FIRST
                                            if uniform_byte_order == ByteOrder.HIGH_ORDER_FIRST
                                            else const.TYPES_LOW_ORDER_FIRST
                                        ),
                                    )
                                    if uniform_byte_order
                                    else If(
                                        [
                                            (
                                                In(
                                                    Variable("Fld"),
                                                    ChoiceList(
                                                        *[
                                                            Variable(f.affixed_name)
                                                            for f, b in message.byte_order.items()
                                                            if b == ByteOrder.LOW_ORDER_FIRST
                                                        ],
                                                    ),
                                                ),
                                                Variable(const.TYPES_LOW_ORDER_FIRST),
                                            ),
                                        ],
                                        Variable(const.TYPES_HIGH_ORDER_FIRST),
                                    )
                                ),
                            ],
                        ),
                    ],
                    [
                        Precondition(
                            AndThen(
                                *self.setter_preconditions(message, "Fld"),
                                In(
                                    Variable("Fld"),
                                    ChoiceList(*[Variable(f.affixed_name) for f in scalar_fields]),
                                ),
                                Call(
                                    self.prefix * message.identifier * "Valid_Value",
                                    [Variable("Fld"), Variable("Val")],
                                ),
                                Call(
                                    self.prefix * message.identifier * "Valid_Size",
                                    [
                                        Variable("Ctx"),
                                        Variable("Fld"),
                                        Call(
                                            self.prefix * message.identifier * "Field_Size",
                                            [Variable("Ctx"), Variable("Fld")],
                                        ),
                                    ],
                                ),
                                common.sufficient_space_for_field_condition(
                                    self.prefix * message.identifier,
                                    Variable("Fld"),
                                ),
                                In(
                                    Call(
                                        self.prefix * message.identifier * "Field_Size",
                                        [Variable("Ctx"), Variable("Fld")],
                                    ),
                                    ValueRange(Number(1), Size(const.TYPES_BASE_INT)),
                                ),
                                Call(
                                    const.TYPES * "Fits_Into",
                                    [
                                        Variable("Val"),
                                        Call(
                                            "Natural",
                                            [
                                                Call(
                                                    self.prefix * message.identifier * "Field_Size",
                                                    [Variable("Ctx"), Variable("Fld")],
                                                ),
                                            ],
                                        ),
                                    ],
                                ),
                            ),
                        ),
                        Postcondition(
                            And(
                                Call("Has_Buffer", [Variable("Ctx")]),
                                Call(
                                    "Valid",
                                    [Variable("Ctx"), Variable("Fld")],
                                ),
                                *(
                                    [
                                        Call(
                                            "Invalid_Successor",
                                            [Variable("Ctx"), Variable("Fld")],
                                        ),
                                    ]
                                    if len(message.fields) > 1
                                    else []
                                ),
                                self.scalar_setter_and_getter_relation(message),
                                *(
                                    [
                                        common.unchanged_cursor_before_or_invalid(
                                            Variable("Fld"),
                                            loop_entry=False,
                                            or_invalid=False,
                                        ),
                                    ]
                                    if len(message.fields) > 1
                                    else []
                                ),
                                *common.context_invariant(message),
                                *[
                                    Equal(e, Old(e))
                                    for e in [
                                        Call("Has_Buffer", [Variable("Ctx")]),
                                        Call("Field_First", [Variable("Ctx"), Variable("Fld")]),
                                    ]
                                ],
                            ),
                        ),
                    ],
                ),
                *[body(f, t) for f, t in scalar_fields.items()],
                *[
                    body(f, t, use_enum_records_directly=True)
                    for f, t in scalar_fields.items()
                    if message.is_definite and isinstance(t, Enumeration) and t.always_valid
                ],
            ],
        )

    def create_composite_setter_empty_procedures(self, message: Message) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Set_{field.name}_Empty",
                [InOutParameter(["Ctx"], "Context")],
            )

        return UnitPart(
            [
                *[
                    SubprogramDeclaration(
                        specification(f),
                        [
                            Precondition(
                                AndThen(
                                    *self.setter_preconditions(
                                        message,
                                        self.prefix * message.identifier * f.affixed_name,
                                    ),
                                    *self.composite_setter_preconditions(message, f),
                                    *self.composite_setter_field_condition_precondition(
                                        message,
                                        f,
                                        empty=True,
                                    ),
                                    Equal(
                                        Call(
                                            self.prefix * message.identifier * "Field_Size",
                                            [
                                                Variable("Ctx"),
                                                Variable(
                                                    self.prefix
                                                    * message.identifier
                                                    * f.affixed_name,
                                                ),
                                            ],
                                        ),
                                        Number(0),
                                    ),
                                ),
                            ),
                            Postcondition(
                                And(
                                    *self.composite_setter_postconditions(f),
                                    *self.public_setter_postconditions(message, f),
                                ),
                            ),
                        ],
                    )
                    for f, t in message.field_types.items()
                    if message.is_possibly_empty(f)
                ],
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        ObjectDeclaration(
                            ["Unused_Buffer_First", "Unused_Buffer_Last"],
                            const.TYPES_INDEX,
                        ),
                        ObjectDeclaration(["Unused_Offset"], const.TYPES_OFFSET),
                    ],
                    [
                        CallStatement(
                            "Set",
                            [
                                Variable("Ctx"),
                                Variable(f.affixed_name),
                                Number(0),
                                Number(0),
                                TRUE,
                                Variable("Unused_Buffer_First"),
                                Variable("Unused_Buffer_Last"),
                                Variable("Unused_Offset"),
                            ],
                        ),
                    ],
                )
                for f, t in message.field_types.items()
                if message.is_possibly_empty(f)
            ],
        )

    def create_sequence_setter_procedures(
        self,
        message: Message,
        sequence_fields: Mapping[Field, Type],
    ) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Set_{field.name}",
                [
                    InOutParameter(["Ctx"], "Context"),
                    Parameter(
                        ["Seq_Ctx"],
                        self.prefix * common.sequence_name(message, field) * "Context",
                    ),
                ],
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Precondition(
                            AndThen(
                                *self.setter_preconditions(
                                    message,
                                    self.prefix * message.identifier * f.affixed_name,
                                ),
                                *self.composite_setter_preconditions(message, f),
                                *self.composite_setter_field_condition_precondition(
                                    message,
                                    f,
                                    empty=True,
                                ),
                                Call(
                                    self.prefix * message.identifier * "Valid_Length",
                                    [
                                        Variable("Ctx"),
                                        Variable(self.prefix * message.identifier * f.affixed_name),
                                        Call(
                                            self.prefix
                                            * common.sequence_name(message, f)
                                            * "Byte_Size",
                                            [Variable("Seq_Ctx")],
                                        ),
                                    ],
                                ),
                                Call(
                                    self.prefix * common.sequence_name(message, f) * "Has_Buffer",
                                    [Variable("Seq_Ctx")],
                                ),
                                Call(
                                    self.prefix * common.sequence_name(message, f) * "Valid",
                                    [Variable("Seq_Ctx")],
                                ),
                            ),
                        ),
                        Postcondition(
                            And(
                                *self.composite_setter_postconditions(f),
                                *self.public_setter_postconditions(message, f),
                                If(
                                    [
                                        (
                                            Greater(
                                                Call(
                                                    "Field_Size",
                                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                                ),
                                                Number(0),
                                            ),
                                            Call(
                                                "Present",
                                                [Variable("Ctx"), Variable(f.affixed_name)],
                                            ),
                                        ),
                                    ],
                                ),
                            ),
                        ),
                    ],
                )
                for f, t in sequence_fields.items()
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        ObjectDeclaration(
                            ["Size"],
                            const.TYPES_BIT_LENGTH,
                            Call(
                                const.TYPES_TO_BIT_LENGTH,
                                [
                                    Call(
                                        self.prefix
                                        * common.sequence_name(message, f)
                                        * "Byte_Size",
                                        [Variable("Seq_Ctx")],
                                    ),
                                ],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(["Unused_First", "Unused_Last"], const.TYPES_BIT_INDEX),
                        ObjectDeclaration(["Buffer_First", "Buffer_Last"], const.TYPES_INDEX),
                        ObjectDeclaration(["Unused_Offset"], const.TYPES_OFFSET),
                    ],
                    [
                        CallStatement(
                            "Set",
                            [
                                Variable("Ctx"),
                                Variable(f.affixed_name),
                                Number(0),
                                Variable("Size"),
                                TRUE,
                                Variable("Buffer_First"),
                                Variable("Buffer_Last"),
                                Variable("Unused_Offset"),
                            ],
                        ),
                        CallStatement(
                            self.prefix * common.sequence_name(message, f) * "Copy",
                            [
                                Variable("Seq_Ctx"),
                                Indexed(
                                    Variable("Ctx.Buffer.all"),
                                    ValueRange(Variable("Buffer_First"), Variable("Buffer_Last")),
                                ),
                            ],
                        ),
                    ],
                )
                for f in sequence_fields
            ],
        )

    def create_opaque_setter_procedures(self, message: Message) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Set_{field.name}",
                [InOutParameter(["Ctx"], "Context"), Parameter(["Data"], const.TYPES_BYTES)],
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Precondition(
                            AndThen(
                                *self.setter_preconditions(
                                    message,
                                    self.prefix * message.identifier * f.affixed_name,
                                ),
                                *self.composite_setter_preconditions(message, f),
                                Call(
                                    self.prefix * message.identifier * "Valid_Length",
                                    [
                                        Variable("Ctx"),
                                        Variable(self.prefix * message.identifier * f.affixed_name),
                                        Length("Data"),
                                    ],
                                ),
                                common.sufficient_space_for_field_condition(
                                    self.prefix * message.identifier,
                                    Variable(self.prefix * message.identifier * f.affixed_name),
                                    Mul(Length("Data"), Size(const.TYPES_BYTE)),
                                ),
                                *self.composite_setter_field_condition_precondition(message, f),
                            ),
                        ),
                        Postcondition(
                            And(
                                *self.composite_setter_postconditions(f),
                                *self.public_setter_postconditions(message, f),
                                Call(
                                    "Equal",
                                    [
                                        Variable("Ctx"),
                                        Variable(f.affixed_name),
                                        Variable("Data"),
                                    ],
                                ),
                            ),
                        ),
                    ],
                )
                for f, t in message.field_types.items()
                if isinstance(t, Opaque)
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        ObjectDeclaration(
                            ["Buffer_First"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [Call("Field_First", [Variable("Ctx"), Variable(f.affixed_name)])],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Buffer_Last"],
                            const.TYPES_INDEX,
                            Add(Variable("Buffer_First"), Length("Data"), -Number(1)),
                            constant=True,
                        ),
                    ],
                    [
                        CallStatement(
                            f"Initialize_{f.name}_Private",
                            [Variable("Ctx"), Length("Data")],
                        ),
                        PragmaStatement(
                            "Assert",
                            [
                                Equal(
                                    Variable("Buffer_Last"),
                                    Call(
                                        const.TYPES_TO_INDEX,
                                        [
                                            Call(
                                                "Field_Last",
                                                [Variable("Ctx"), Variable(f.affixed_name)],
                                            ),
                                        ],
                                    ),
                                ),
                            ],
                        ),
                        Assignment(
                            Slice(
                                Selected(Variable("Ctx.Buffer"), "all"),
                                Variable("Buffer_First"),
                                Variable("Buffer_Last"),
                            ),
                            Variable("Data"),
                        ),
                        # Improve provability of `Equal` in postcondition
                        PragmaStatement(
                            "Assert",
                            [
                                Equal(
                                    Slice(
                                        Selected(Variable("Ctx.Buffer"), "all"),
                                        Call(
                                            const.TYPES_TO_INDEX,
                                            [
                                                Call(
                                                    "Field_First",
                                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                                ),
                                            ],
                                        ),
                                        Call(
                                            const.TYPES_TO_INDEX,
                                            [
                                                Call(
                                                    "Field_Last",
                                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                                ),
                                            ],
                                        ),
                                    ),
                                    Variable("Data"),
                                ),
                            ],
                        ),
                    ],
                )
                for f, t in message.field_types.items()
                if isinstance(t, Opaque)
            ],
        )

    def create_generic_opaque_setter_procedures(self, message: Message) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Generic_Set_{field.name}",
                [InOutParameter(["Ctx"], "Context"), Parameter(["Length"], const.TYPES_LENGTH)],
            )

        def formal_parameters(field: Field) -> list[FormalSubprogramDeclaration]:
            return [
                FormalSubprogramDeclaration(
                    ProcedureSpecification(
                        f"Process_{field.name}",
                        [
                            OutParameter([field.name], const.TYPES_BYTES),
                        ],
                    ),
                ),
                FormalSubprogramDeclaration(
                    FunctionSpecification(
                        "Process_Data_Pre",
                        "Boolean",
                        [Parameter(["Length"], const.TYPES_LENGTH)],
                    ),
                ),
            ]

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Precondition(
                            AndThen(
                                *self.setter_preconditions(
                                    message,
                                    self.prefix * message.identifier * f.affixed_name,
                                ),
                                *self.composite_setter_preconditions(message, f),
                                Call(
                                    self.prefix * message.identifier * "Valid_Length",
                                    [
                                        Variable("Ctx"),
                                        Variable(self.prefix * message.identifier * f.affixed_name),
                                        Variable("Length"),
                                    ],
                                ),
                                GreaterEqual(
                                    Call(
                                        const.TYPES_TO_LENGTH,
                                        [
                                            Call(
                                                self.prefix
                                                * message.identifier
                                                * "Available_Space",
                                                [
                                                    Variable("Ctx"),
                                                    Variable(
                                                        self.prefix
                                                        * message.identifier
                                                        * f.affixed_name,
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                    Variable("Length"),
                                ),
                                Call(
                                    "Process_Data_Pre",
                                    [
                                        Variable("Length"),
                                    ],
                                ),
                            ),
                        ),
                        Postcondition(
                            And(
                                *self.composite_setter_postconditions(f),
                                *self.public_setter_postconditions(message, f),
                            ),
                        ),
                    ],
                    formal_parameters(f),
                )
                for f, t in message.field_types.items()
                if isinstance(t, Opaque)
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        ObjectDeclaration(
                            ["First"],
                            const.TYPES_BIT_INDEX,
                            Call("Field_First", [Variable("Ctx"), Variable(f.affixed_name)]),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Buffer_First"],
                            const.TYPES_INDEX,
                            Call(const.TYPES_TO_INDEX, [Variable("First")]),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Buffer_Last"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Add(
                                        Variable("First"),
                                        Call(const.TYPES_TO_BIT_LENGTH, [Variable("Length")]),
                                        -Number(1),
                                    ),
                                ],
                            ),
                            constant=True,
                        ),
                    ],
                    [
                        CallStatement(
                            f"Process_{f.name}",
                            [
                                Slice(
                                    Selected(Variable("Ctx.Buffer"), "all"),
                                    Variable("Buffer_First"),
                                    Variable("Buffer_Last"),
                                ),
                            ],
                        ),
                        CallStatement(
                            f"Initialize_{f.name}_Private",
                            [
                                Variable("Ctx"),
                                Variable("Length"),
                            ],
                        ),
                    ],
                )
                for f, t in message.field_types.items()
                if isinstance(t, Opaque)
            ],
        )

    def create_composite_initialize_procedures(
        self,
        message: Message,
        fields_with_explicit_size: list[Field],
        fields_with_implicit_size: list[Field],
    ) -> UnitPart:
        def specification_private(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Initialize_{field.name}_Private",
                [InOutParameter(["Ctx"], "Context"), Parameter(["Length"], const.TYPES_LENGTH)],
            )

        def specification_public(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Initialize_{field.name}",
                [InOutParameter(["Ctx"], "Context")],
            )

        def specification_public_with_length(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Initialize_{field.name}",
                [InOutParameter(["Ctx"], "Context"), Parameter(["Length"], const.TYPES_LENGTH)],
            )

        return UnitPart(
            [
                s
                for f, t in message.field_types.items()
                if isinstance(t, (Opaque, Sequence))
                for s in [
                    *(
                        [
                            SubprogramDeclaration(
                                specification_public(f),
                                [
                                    Precondition(
                                        AndThen(
                                            *self.setter_preconditions(
                                                message,
                                                self.prefix * message.identifier * f.affixed_name,
                                            ),
                                            *self.composite_setter_preconditions(message, f),
                                        ),
                                    ),
                                    Postcondition(
                                        AndThen(
                                            *self.composite_setter_postconditions(f),
                                            *self.public_setter_postconditions(message, f),
                                        ),
                                    ),
                                ],
                            ),
                        ]
                        if f in fields_with_explicit_size
                        else []
                    ),
                    *(
                        [
                            SubprogramDeclaration(
                                specification_public_with_length(f),
                                [
                                    Precondition(
                                        AndThen(
                                            *self.setter_preconditions(
                                                message,
                                                self.prefix * message.identifier * f.affixed_name,
                                            ),
                                            Call(
                                                self.prefix * message.identifier * "Valid_Length",
                                                [
                                                    Variable("Ctx"),
                                                    Variable(
                                                        self.prefix
                                                        * message.identifier
                                                        * f.affixed_name,
                                                    ),
                                                    Variable("Length"),
                                                ],
                                            ),
                                            *self.composite_setter_preconditions(
                                                message,
                                                f,
                                                Call(
                                                    const.TYPES_TO_BIT_LENGTH,
                                                    [Variable("Length")],
                                                ),
                                            ),
                                        ),
                                    ),
                                    Postcondition(
                                        AndThen(
                                            *self.composite_setter_postconditions(f),
                                            Equal(
                                                Call(
                                                    "Field_Size",
                                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                                ),
                                                Call(
                                                    const.TYPES_TO_BIT_LENGTH,
                                                    [Variable("Length")],
                                                ),
                                            ),
                                            *self.public_setter_postconditions(message, f),
                                        ),
                                    ),
                                ],
                            ),
                        ]
                        if f in fields_with_implicit_size
                        else []
                    ),
                ]
            ],
            [
                s
                for f, t in message.field_types.items()
                if isinstance(t, (Opaque, Sequence))
                for s in [
                    SubprogramBody(
                        specification_private(f),
                        [
                            ObjectDeclaration(
                                ["First"],
                                const.TYPES_BIT_INDEX,
                                Call("Field_First", [Variable("Ctx"), Variable(f.affixed_name)]),
                                constant=True,
                            ),
                            ObjectDeclaration(
                                ["Last"],
                                const.TYPES_BIT_INDEX,
                                Add(
                                    Call(
                                        "Field_First",
                                        [Variable("Ctx"), Variable(f.affixed_name)],
                                    ),
                                    Mul(
                                        Call(const.TYPES_BIT_LENGTH, [Variable("Length")]),
                                        Size(const.TYPES_BYTE),
                                    ),
                                    -Number(1),
                                ),
                                constant=True,
                            ),
                        ],
                        [
                            # Improve provability of context predicate
                            PragmaStatement(
                                "Assert",
                                [Equal(Mod(Variable("Last"), Size(const.TYPES_BYTE)), Number(0))],
                            ),
                            *common.initialize_field_statements(f, reset_written_last=True),
                        ],
                        [
                            Precondition(
                                AndThen(
                                    *self.setter_preconditions(
                                        message,
                                        self.prefix * message.identifier * f.affixed_name,
                                    ),
                                    Call(
                                        self.prefix * message.identifier * "Valid_Length",
                                        [
                                            Variable("Ctx"),
                                            Variable(
                                                self.prefix * message.identifier * f.affixed_name,
                                            ),
                                            Variable("Length"),
                                        ],
                                    ),
                                    GreaterEqual(
                                        Call(
                                            const.TYPES_TO_LENGTH,
                                            [
                                                Call(
                                                    self.prefix
                                                    * message.identifier
                                                    * "Available_Space",
                                                    [
                                                        Variable("Ctx"),
                                                        Variable(
                                                            self.prefix
                                                            * message.identifier
                                                            * f.affixed_name,
                                                        ),
                                                    ],
                                                ),
                                            ],
                                        ),
                                        Variable("Length"),
                                    ),
                                    Equal(
                                        Mod(
                                            Call(
                                                self.prefix * message.identifier * "Field_First",
                                                [
                                                    Variable("Ctx"),
                                                    Variable(
                                                        self.prefix
                                                        * message.identifier
                                                        * f.affixed_name,
                                                    ),
                                                ],
                                            ),
                                            Size(const.TYPES_BYTE),
                                        ),
                                        Number(1),
                                    ),
                                ),
                            ),
                            Postcondition(
                                AndThen(
                                    *self.composite_setter_postconditions(f),
                                    Equal(
                                        Call(
                                            "Field_Size",
                                            [Variable("Ctx"), Variable(f.affixed_name)],
                                        ),
                                        Call(const.TYPES_TO_BIT_LENGTH, [Variable("Length")]),
                                    ),
                                    *self.private_setter_postconditions(message, f),
                                ),
                            ),
                        ],
                    ),
                    *(
                        [
                            SubprogramBody(
                                specification_public(f),
                                [],
                                [
                                    CallStatement(
                                        specification_private(f).identifier,
                                        [
                                            Variable("Ctx"),
                                            Call(
                                                const.TYPES_TO_LENGTH,
                                                [
                                                    Call(
                                                        "Field_Size",
                                                        [Variable("Ctx"), Variable(f.affixed_name)],
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        ]
                        if f in fields_with_explicit_size
                        else []
                    ),
                    *(
                        [
                            SubprogramBody(
                                specification_public_with_length(f),
                                [],
                                [
                                    CallStatement(
                                        specification_private(f).identifier,
                                        [Variable("Ctx"), Variable("Length")],
                                    ),
                                ],
                            ),
                        ]
                        if f in fields_with_implicit_size
                        else []
                    ),
                ]
            ],
        )

    def setter_preconditions(self, message: Message, field_name: StrID) -> list[Expr]:
        return [
            Not(Constrained("Ctx")),
            Call(self.prefix * message.identifier * "Has_Buffer", [Variable("Ctx")]),
            Call(
                self.prefix * message.identifier * "Valid_Next",
                [Variable("Ctx"), Variable(field_name)],
            ),
        ]

    def private_setter_postconditions(self, message: Message, field: Field) -> list[Expr]:
        return [
            Equal(
                Variable("Ctx.Verified_Last"),
                Call(
                    "Field_Last",
                    [Variable("Ctx"), Variable(field.affixed_name)],
                ),
            ),
            *self.setter_postconditions(message, field),
        ]

    def public_setter_postconditions(self, message: Message, field: Field) -> list[Expr]:
        return [
            *(
                [
                    If(
                        [
                            (
                                Call("Well_Formed_Message", [Variable("Ctx")]),
                                Equal(
                                    Call("Message_Last", [Variable("Ctx")]),
                                    Call(
                                        "Field_Last",
                                        [Variable("Ctx"), Variable(field.affixed_name)],
                                    ),
                                ),
                            ),
                        ],
                    ),
                ]
                if field in message.direct_predecessors(FINAL)
                else []
            ),
            *self.setter_postconditions(message, field),
        ]

    def setter_postconditions(self, message: Message, field: Field) -> list[Expr]:
        return [
            *[
                Call("Invalid", [Variable("Ctx"), Variable(p.affixed_name)])
                for p in message.successors(field)
                if p != FINAL
            ],
            *common.valid_path_to_next_field_condition(message, field, self.prefix),
            *common.context_invariant(message),
            *[
                Equal(e, Old(e))
                for e in [
                    Call(
                        "Valid_Next",
                        [Variable("Ctx"), Variable(field.affixed_name)],
                    ),
                ]
                + [
                    Call(f"Get_{p.name}", [Variable("Ctx")])
                    for p in message.definite_predecessors(field)
                    for t in [message.field_types[p]]
                    if isinstance(t, Scalar) and int(t.value_count) > 1
                ]
                + [Call("Field_First", [Variable("Ctx"), Variable(field.affixed_name)])]
            ],
        ]

    @staticmethod
    def composite_setter_postconditions(field: Field) -> list[Expr]:
        return [
            Call("Has_Buffer", [Variable("Ctx")]),
            Call("Well_Formed", [Variable("Ctx"), Variable(field.affixed_name)]),
        ]

    def composite_setter_field_condition_precondition(
        self,
        message: Message,
        field: Field,
        empty: bool = False,
    ) -> list[Expr]:
        return [
            common.field_condition_call(
                self.prefix,
                message,
                field,
                aggregate=None if empty else Variable("Data"),
                size=None if empty else Call(const.TYPES_TO_BIT_LENGTH, [Length("Data")]),
            ),
        ]

    def composite_setter_preconditions(
        self,
        message: Message,
        field: Field,
        size: Optional[Expr] = None,
    ) -> list[Expr]:
        return [
            common.sufficient_space_for_field_condition(
                self.prefix * message.identifier,
                Variable(self.prefix * message.identifier * field.affixed_name),
                size,
            ),
        ]

    def scalar_setter_and_getter_relation(
        self,
        message: Message,
    ) -> Expr:
        return AndThen(
            Equal(
                Selected(
                    Indexed(
                        Selected(Variable("Ctx"), "Cursors"),
                        Variable("Fld"),
                    ),
                    "Value",
                ),
                Variable("Val"),
            ),
            If(
                [
                    (
                        AndThen(
                            In(
                                Variable("Fld"),
                                ChoiceList(
                                    *[
                                        Variable(f.affixed_name)
                                        for f in message.direct_predecessors(FINAL)
                                    ],
                                ),
                            ),
                            Call(
                                "Well_Formed_Message",
                                [Variable("Ctx")],
                            ),
                        ),
                        Equal(
                            Call(
                                "Message_Last",
                                [Variable("Ctx")],
                            ),
                            Call(
                                "Field_Last",
                                [
                                    Variable("Ctx"),
                                    Variable("Fld"),
                                ],
                            ),
                        ),
                    ),
                ],
            ),
        )

    @staticmethod
    def _update_last(
        message: Optional[Message] = None,
        field: Optional[Field] = None,
    ) -> list[Statement]:
        assert (message and field) or not (message or field)
        last = (
            Variable("Last")
            if message and field and field in message.direct_predecessors(FINAL)
            else Mul(
                Div(
                    Add(
                        Variable("Last"),
                        Size(const.TYPES_BYTE),
                        -Number(1),
                    ),
                    Size(const.TYPES_BYTE),
                ),
                Size(const.TYPES_BYTE),
            )
        )
        return [
            # Improve provability of context predicate
            PragmaStatement("Assert", [Equal(Mod(last, Size(const.TYPES_BYTE)), Number(0))]),
            # Eng/RecordFlux/RecordFlux#868
            PragmaStatement(
                "Warnings",
                [
                    Variable("Off"),
                    String("attribute Update is an obsolescent feature"),
                ],
            ),
            Assignment(
                "Ctx",
                Update(
                    "Ctx",
                    ("Verified_Last", last),
                    ("Written_Last", last),
                ),
            ),
            PragmaStatement(
                "Warnings",
                [
                    Variable("On"),
                    String("attribute Update is an obsolescent feature"),
                ],
            ),
        ]



from __future__ import annotations

import typing as ty

from pydantic import BaseModel

# TODO(eng/recordflux/RecordFlux#1359): Replace ty.* by collections.abc.*
# Sequence and Mapping are imported from collections.abc as importing them
# from typing is deprecated. However pydantic does not support the imported
# version from collections.abc. To fix that typing is imported as ty and the
# typing versions of Sequence and Mapping are used in classes that derive
# from pydantic.BaseModel.
# This is only relevant for Python 3.8.


class ProofAttempt(BaseModel):  # type: ignore[misc]
    result: str
    steps: int
    time: float


class CheckTreeElement(BaseModel):  # type: ignore[misc]
    proof_attempts: ty.Mapping[str, ProofAttempt]


class Proof(BaseModel):  # type: ignore[misc]
    file: str
    line: int
    col: int
    rule: str
    check_tree: ty.Sequence[CheckTreeElement]


class SPARKFile(BaseModel):  # type: ignore[misc]
    proof: ty.Sequence[Proof]



from __future__ import annotations

import enum
import math
import textwrap
from collections.abc import Callable
from typing import Optional

from rflx import expression as expr, model
from rflx.ada import (
    TRUE,
    Add,
    And,
    AndThen,
    Assignment,
    Call,
    CallStatement,
    ContextItem,
    Declaration,
    Equal,
    Expr,
    First,
    ForAllIn,
    GenericPackageInstantiation,
    GreaterEqual,
    If,
    IfExpr,
    Indexed,
    Last,
    Less,
    LessEqual,
    LoopEntry,
    Max,
    Name,
    NamedAggregate,
    NotEqual,
    Number,
    ObjectDeclaration,
    Old,
    Parameter,
    Pragma,
    PragmaStatement,
    Rem,
    Selected,
    Size,
    Statement,
    String,
    Sub,
    Update,
    ValueRange,
    Variable,
    WithClause,
)
from rflx.const import BUILTINS_PACKAGE
from rflx.identifier import ID

from . import const

EMPTY_ARRAY = NamedAggregate((ValueRange(Number(1), Number(0)), Number(0)))


class Debug(enum.Enum):
    NONE = enum.auto()
    BUILTIN = enum.auto()
    EXTERNAL = enum.auto()


def substitution(
    message: model.Message,
    prefix: str,
    embedded: bool = False,
    public: bool = False,
    target_type: Optional[ID] = const.TYPES_BASE_INT,
) -> Callable[[expr.Expr], expr.Expr]:
    facts = substitution_facts(message, prefix, embedded, public, target_type)

    def type_conversion(expression: expr.Expr) -> expr.Expr:
        return expr.Call(target_type, [expression]) if target_type else expression

    def func(  # noqa: PLR0912
        expression: expr.Expr,
    ) -> expr.Expr:
        def byte_aggregate(aggregate: expr.Aggregate) -> expr.Aggregate:
            return expr.Aggregate(*[expr.Val(const.TYPES_BYTE, e) for e in aggregate.elements])

        if isinstance(expression, expr.Name) and expression in facts:
            return facts[expression]

        if isinstance(expression, expr.String):
            return expr.Aggregate(*expression.elements)

        if isinstance(expression, expr.And):
            return expr.AndThen(*[func(t) for t in expression.terms])

        if isinstance(expression, expr.Or):
            return expr.OrElse(*[func(t) for t in expression.terms])

        if isinstance(expression, (expr.Equal, expr.NotEqual)):
            field = None
            aggregate = None
            if isinstance(expression.left, expr.Variable) and isinstance(
                expression.right,
                expr.Aggregate,
            ):
                field = model.Field(expression.left.name)
                aggregate = byte_aggregate(expression.right)
            elif isinstance(expression.left, expr.Aggregate) and isinstance(
                expression.right,
                expr.Variable,
            ):
                field = model.Field(expression.right.name)
                aggregate = byte_aggregate(expression.left)
            if field and field in message.fields and len(field.identifier.parts) == 1 and aggregate:
                if embedded:
                    return expression.__class__(
                        expr.Indexed(
                            expr.Variable(ID("Buffer") * "all"),
                            expr.ValueRange(
                                expr.Call(
                                    const.TYPES_TO_INDEX,
                                    [
                                        expr.Selected(
                                            expr.Indexed(
                                                expr.Variable("Cursors"),
                                                expr.Variable(field.affixed_name),
                                            ),
                                            "First",
                                        ),
                                    ],
                                ),
                                expr.Call(
                                    const.TYPES_TO_INDEX,
                                    [
                                        expr.Selected(
                                            expr.Indexed(
                                                expr.Variable("Cursors"),
                                                expr.Variable(field.affixed_name),
                                            ),
                                            "Last",
                                        ),
                                    ],
                                ),
                            ),
                        ),
                        aggregate,
                    )
                equal_call = expr.Call(
                    "Equal",
                    [expr.Variable("Ctx"), expr.Variable(field.affixed_name), aggregate],
                )
                return equal_call if isinstance(expression, expr.Equal) else expr.Not(equal_call)

            boolean_literal = None
            other = None
            if (
                isinstance(expression.left, expr.Literal)
                and expression.left.identifier in model.BOOLEAN.literals
            ):
                boolean_literal = expression.left
                other = expression.right
            if (
                isinstance(expression.right, expr.Literal)
                and expression.right.identifier in model.BOOLEAN.literals
            ):
                boolean_literal = expression.right
                other = expression.left
            if boolean_literal and other:
                return expression.__class__(
                    other,
                    type_conversion(expr.Call("To_Base_Integer", [boolean_literal])),
                )

        def field_value(field: model.Field) -> expr.Expr:
            if public:
                return expr.Call(f"Get_{field.name}", [expr.Variable("Ctx")])
            return expr.Selected(
                expr.Indexed(
                    expr.Variable(ID("Ctx") * "Cursors" if not embedded else "Cursors"),
                    expr.Variable(field.affixed_name),
                ),
                "Value",
            )

        if isinstance(expression, expr.Relation):
            if (
                isinstance(expression.left, expr.Variable)
                and model.Field(expression.left.name) in message.fields
                and isinstance(expression.right, expr.Number)
            ):
                return expression.__class__(
                    field_value(model.Field(expression.left.name)),
                    expression.right,
                )
            if (
                isinstance(expression.right, expr.Variable)
                and model.Field(expression.right.name) in message.fields
                and isinstance(expression.left, expr.Number)
            ):
                return expression.__class__(
                    expression.left,
                    field_value(model.Field(expression.right.name)),
                )

        return expression

    return func


def substitution_facts(
    message: model.Message,
    prefix: str,
    embedded: bool = False,
    public: bool = False,
    target_type: Optional[ID] = const.TYPES_BASE_INT,
) -> dict[expr.Name, expr.Expr]:
    def prefixed(name: str) -> expr.Expr:
        return expr.Variable(ID("Ctx") * name) if not embedded else expr.Variable(name)

    first = prefixed("First")
    last = expr.Call("Written_Last", [expr.Variable("Ctx")]) if public else prefixed("Written_Last")
    cursors = prefixed("Cursors")

    def field_first(field: model.Field) -> expr.Expr:
        if public:
            return expr.Call(
                "Field_First",
                [expr.Variable("Ctx"), expr.Variable(field.affixed_name)],
            )
        return expr.Selected(expr.Indexed(cursors, expr.Variable(field.affixed_name)), "First")

    def field_last(field: model.Field) -> expr.Expr:
        if public:
            return expr.Call(
                "Field_Last",
                [expr.Variable("Ctx"), expr.Variable(field.affixed_name)],
            )
        return expr.Selected(expr.Indexed(cursors, expr.Variable(field.affixed_name)), "Last")

    def field_size(field: model.Field) -> expr.Expr:
        if public:
            return expr.Call(
                "Field_Size",
                [expr.Variable("Ctx"), expr.Variable(field.affixed_name)],
            )
        return expr.Add(
            expr.Sub(
                expr.Selected(expr.Indexed(cursors, expr.Variable(field.affixed_name)), "Last"),
                expr.Selected(expr.Indexed(cursors, expr.Variable(field.affixed_name)), "First"),
            ),
            expr.Number(1),
        )

    def parameter_value(parameter: model.Field, parameter_type: model.Type) -> expr.Expr:
        if isinstance(parameter_type, model.Enumeration):
            if embedded:
                return expr.Call("To_Base_Integer", [expr.Variable(parameter.name)])
            return expr.Call("To_Base_Integer", [expr.Variable("Ctx" * parameter.identifier)])
        if isinstance(parameter_type, model.Scalar):
            if embedded:
                return expr.Variable(parameter.name)
            return expr.Variable("Ctx" * parameter.identifier)

        assert False, f'unexpected type "{type(parameter_type).__name__}"'

    def field_value(field: model.Field, field_type: model.Type) -> expr.Expr:
        if isinstance(field_type, model.Enumeration):
            if public:
                return expr.Call(
                    "To_Base_Integer",
                    [expr.Call(f"Get_{field.name}", [expr.Variable("Ctx")])],
                )
            return expr.Selected(
                expr.Indexed(cursors, expr.Variable(field.affixed_name)),
                "Value",
            )
        if isinstance(field_type, model.Scalar):
            if public:
                return expr.Call(f"Get_{field.name}", [expr.Variable("Ctx")])
            return expr.Selected(
                expr.Indexed(cursors, expr.Variable(field.affixed_name)),
                "Value",
            )
        if isinstance(field_type, model.Composite):
            return expr.Variable(field.name)

        assert False, f'unexpected type "{type(field_type).__name__}"'

    def type_conversion(expression: expr.Expr) -> expr.Expr:
        return expr.Call(target_type, [expression]) if target_type else expression

    return {
        expr.First("Message"): type_conversion(first),
        expr.Last("Message"): type_conversion(last),
        expr.Size("Message"): type_conversion(expr.Add(last, -first, expr.Number(1))),
        **{expr.First(f.name): type_conversion(field_first(f)) for f in message.fields},
        **{expr.Last(f.name): type_conversion(field_last(f)) for f in message.fields},
        **{expr.Size(f.name): type_conversion(field_size(f)) for f in message.fields},
        **{
            expr.Variable(p.identifier): type_conversion(parameter_value(p, t))
            for p, t in message.parameter_types.items()
        },
        **{
            expr.Variable(f.name): type_conversion(field_value(f, t))
            for f, t in message.field_types.items()
        },
        **{
            expr.Literal(l): type_conversion(expr.Call("To_Base_Integer", [expr.Variable(l)]))
            for t in message.types.values()
            if isinstance(t, model.Enumeration) and t != model.BOOLEAN
            for l in t.literals
        },
        **{
            expr.Literal(t.package * l): type_conversion(
                expr.Call("To_Base_Integer", [expr.Variable(prefix * t.package * l)]),
            )
            for t in message.types.values()
            if isinstance(t, model.Enumeration) and t != model.BOOLEAN
            for l in t.literals
        },
        # Eng/RecordFlux/RecordFlux#276
        **{expr.ValidChecksum(f): expr.TRUE for f in message.checksums},
    }


def message_structure_invariant(
    message: model.Message,
    prefix: str,
    embedded: bool = False,
) -> Expr:
    """
    Create the invariant that defines a valid message structure.

    This invariant ensures the properties of message fields that can depend on the concrete message
    path: the field size, the location of the field and the predecessor of the field. This is
    realized by a list of if-expressions. Each if-expression states the properties of a given field,
    based on the incoming links. Simplifications are applied (e.g. when there is only one
    incoming link).
    """

    def prefixed(name: str) -> expr.Expr:
        return expr.Selected(expr.Variable("Ctx"), name) if not embedded else expr.Variable(name)

    def link_property(link: model.Link, unique: bool) -> Expr:
        field_type = message.types[link.target]
        condition = link.condition.substituted(substitution(message, prefix, embedded)).simplified()
        size = (
            field_type.size
            if isinstance(field_type, model.Scalar)
            else link.size.substituted(
                substitution(message, prefix, embedded, target_type=const.TYPES_BIT_LENGTH),
            ).simplified()
        )
        first = (
            prefixed("First")
            if link.source == model.INITIAL
            else link.first.substituted(
                substitution(message, prefix, embedded, target_type=const.TYPES_BIT_INDEX),
            )
            .substituted(
                mapping={
                    expr.UNDEFINED: expr.Add(
                        expr.Selected(
                            expr.Indexed(
                                prefixed("Cursors"),
                                expr.Variable(link.source.affixed_name),
                            ),
                            "Last",
                        ),
                        expr.Number(1),
                    ),
                },
            )
            .simplified()
        )
        precond = (
            AndThen(
                Call(
                    "Well_Formed",
                    [Indexed(Variable("Cursors"), Variable(link.source.affixed_name))],
                ),
                condition.ada_expr(),
            )
            if link.source != model.INITIAL and not unique
            else TRUE
        )

        return If(
            [
                (
                    precond,
                    AndThen(
                        Equal(
                            Add(
                                Sub(
                                    Selected(
                                        Indexed(
                                            prefixed("Cursors").ada_expr(),
                                            Variable(link.target.affixed_name),
                                        ),
                                        "Last",
                                    ),
                                    Selected(
                                        Indexed(
                                            prefixed("Cursors").ada_expr(),
                                            Variable(link.target.affixed_name),
                                        ),
                                        "First",
                                    ),
                                ),
                                Number(1),
                            ),
                            size.ada_expr(),
                        ),
                        Equal(
                            Selected(
                                Indexed(
                                    prefixed("Cursors").ada_expr(),
                                    Variable(link.target.affixed_name),
                                ),
                                "First",
                            ),
                            first.ada_expr(),
                        ),
                    ),
                ),
            ],
        )

    def field_property(fld: model.Field) -> Expr:
        incoming = message.incoming(fld)
        unique = len(incoming) == 1
        return AndThen(*[link_property(link, unique) for link in incoming])

    def map_invariant(fld: model.Field) -> Expr:
        return If(
            [
                (
                    Call(
                        "Well_Formed",
                        [Indexed(prefixed("Cursors").ada_expr(), Variable(fld.affixed_name))],
                    ),
                    field_property(fld),
                ),
            ],
        )

    return AndThen(*[map_invariant(f) for f in message.fields])


def context_predicate(
    message: model.Message,
    prefix: str,
) -> Expr:
    def invalid_successors_invariant() -> Expr:
        """
        Create the invariant that defines the state of successors of invalid fields.

        This invariant ensures for all invalid message fields that all its successor fields are also
        invalid.
        """
        return AndThen(
            *[
                If(
                    [
                        (
                            AndThen(
                                *[
                                    Call(
                                        "Invalid",
                                        [
                                            Indexed(
                                                Variable("Cursors"),
                                                Variable(p.affixed_name),
                                            ),
                                        ],
                                    )
                                    for p in message.direct_predecessors(f)
                                ],
                            ),
                            Call(
                                "Invalid",
                                [
                                    Indexed(
                                        Variable("Cursors"),
                                        Variable(f.affixed_name),
                                    ),
                                ],
                            ),
                        ),
                    ],
                )
                for f in message.fields
                if f not in message.direct_successors(model.INITIAL)
            ],
        )

    return AndThen(
        If(
            [
                (
                    NotEqual(Variable("Buffer"), Variable("null")),
                    And(
                        Equal(First("Buffer"), Variable("Buffer_First")),
                        Equal(Last("Buffer"), Variable("Buffer_Last")),
                    ),
                ),
            ],
        ),
        public_context_predicate(),
        LessEqual(Sub(Variable("First"), Number(1)), Variable("Verified_Last")),
        LessEqual(Sub(Variable("First"), Number(1)), Variable("Written_Last")),
        LessEqual(Variable("Verified_Last"), Variable("Written_Last")),
        LessEqual(Variable("Written_Last"), Variable("Last")),
        Equal(Rem(Variable("First"), Size(const.TYPES_BYTE)), Number(1)),
        Equal(Rem(Variable("Last"), Size(const.TYPES_BYTE)), Number(0)),
        Equal(Rem(Variable("Verified_Last"), Size(const.TYPES_BYTE)), Number(0)),
        Equal(Rem(Variable("Written_Last"), Size(const.TYPES_BYTE)), Number(0)),
        Call(
            "Cursors_Invariant",
            [Variable("Cursors"), Variable("First"), Variable("Verified_Last")],
        ),
        Call(
            "Valid_Predecessors_Invariant",
            [
                Variable("Cursors"),
                Variable("First"),
                Variable("Verified_Last"),
                Variable("Written_Last"),
                Variable("Buffer"),
                *[Variable(p.identifier) for p in message.parameter_types],
            ],
        ),
        invalid_successors_invariant(),
        message_structure_invariant(message, prefix, embedded=True),
    )


def public_context_predicate() -> Expr:
    """
    Create the predicate that defines the relations between the public discriminants of the context.

    This invariant ensures that

        - the positions of the first and last usable bit are inside the range of the buffer,
        - the size of the usable buffer is greater or equal to zero,
        - the value of `Last` and `Buffer_Last` is less than the last possible value of their
          respective type to allow the representation of a size of zero in these edge cases
        - and the positions of the first and last usable bit are byte aligned.
    """
    return And(
        GreaterEqual(Call(const.TYPES_TO_INDEX, [Variable("First")]), Variable("Buffer_First")),
        LessEqual(Call(const.TYPES_TO_INDEX, [Variable("Last")]), Variable("Buffer_Last")),
        Less(Variable("Buffer_Last"), Last(const.TYPES_INDEX)),
        LessEqual(Variable("First"), Add(Variable("Last"), Number(1))),
        Less(Variable("Last"), Last(const.TYPES_BIT_INDEX)),
        Equal(Rem(Variable("First"), Size(const.TYPES_BYTE)), Number(1)),
        Equal(Rem(Variable("Last"), Size(const.TYPES_BYTE)), Number(0)),
    )


def context_invariant(message: model.Message, loop_entry: bool = False) -> list[Expr]:
    return [
        Equal(e, LoopEntry(e) if loop_entry else Old(e))
        for e in [
            Variable("Ctx.Buffer_First"),
            Variable("Ctx.Buffer_Last"),
            Variable("Ctx.First"),
            Variable("Ctx.Last"),
            *[Variable("Ctx" * ID(p.name)) for p in message.parameters],
        ]
    ]


def valid_path_to_next_field_condition(
    message: model.Message,
    field: model.Field,
    prefix: str,
) -> list[Expr]:
    return [
        If(
            [
                (
                    l.condition.substituted(substitution(message, public=True, prefix=prefix))
                    .simplified()
                    .ada_expr(),
                    (
                        Call(
                            "Valid_Next",
                            [Variable("Ctx"), Variable(l.target.affixed_name)],
                        )
                        if l.target != model.FINAL
                        else TRUE
                    ),
                ),
            ],
        )
        for l in message.outgoing(field)
        if l.target != model.FINAL
    ]


def context_cursor_unchanged(
    message: model.Message,
    field: model.Field,
    predecessors: bool,
) -> list[Expr]:
    lower: model.Field
    upper: model.Field
    if predecessors:
        if len(message.predecessors(field)) == 0:
            return []
        lower = message.fields[0]
        upper = message.fields[message.fields.index(field) - 1]
    else:
        if len(message.successors(field)) == 0:
            return []
        lower = message.fields[message.fields.index(field) + 1]
        upper = message.fields[-1]
    return [
        ForAllIn(
            "F",
            ValueRange(
                lower=Variable(lower.affixed_name),
                upper=Variable(upper.affixed_name),
                type_identifier=ID("Field"),
            ),
            Equal(
                Call(
                    "Context_Cursors_Index",
                    [
                        Call(
                            "Context_Cursors",
                            [Variable("Ctx")],
                        ),
                        Variable("F"),
                    ],
                ),
                Call(
                    "Context_Cursors_Index",
                    [
                        Old(
                            Call(
                                "Context_Cursors",
                                [Variable("Ctx")],
                            ),
                        ),
                        Variable("F"),
                    ],
                ),
            ),
        ),
    ]


def sufficient_space_for_field_condition(
    message_id: ID,
    field_name: Name,
    size: Optional[Expr] = None,
) -> Expr:
    if size is None:
        size = Call(message_id * "Field_Size", [Variable("Ctx"), field_name])
    return GreaterEqual(Call(message_id * "Available_Space", [Variable("Ctx"), field_name]), size)


def initialize_field_statements(
    field: model.Field,
    reset_written_last: bool = False,
) -> list[Statement]:
    return [
        CallStatement(
            "Reset_Dependent_Fields",
            [Variable("Ctx"), Variable(field.affixed_name)],
        ),
        # Eng/RecordFlux/RecordFlux#868
        PragmaStatement(
            "Warnings",
            [
                Variable("Off"),
                String("attribute Update is an obsolescent feature"),
            ],
        ),
        Assignment(
            "Ctx",
            Update(
                "Ctx",
                ("Verified_Last", Variable("Last")),
                (
                    "Written_Last",
                    (
                        Variable("Last")
                        if reset_written_last
                        else Max(
                            const.TYPES_BIT_LENGTH,
                            Variable("Ctx.Written_Last"),
                            Variable("Last"),
                        )
                    ),
                ),
            ),
        ),
        PragmaStatement(
            "Warnings",
            [
                Variable("On"),
                String("attribute Update is an obsolescent feature"),
            ],
        ),
        Assignment(
            Indexed(Variable("Ctx.Cursors"), Variable(field.affixed_name)),
            NamedAggregate(
                ("State", Variable("S_Well_Formed")),
                ("First", Variable("First")),
                ("Last", Variable("Last")),
                ("Value", Number(0)),
            ),
        ),
    ]


def field_bit_location_declarations(field_name: Name) -> list[Declaration]:
    return [
        ObjectDeclaration(
            ["First"],
            const.TYPES_BIT_INDEX,
            Call("Field_First", [Variable("Ctx"), field_name]),
            constant=True,
        ),
        ObjectDeclaration(
            ["Last"],
            const.TYPES_BIT_INDEX,
            Call("Field_Last", [Variable("Ctx"), field_name]),
            constant=True,
        ),
    ]


def field_condition_call(
    prefix: str,
    message: model.Message,
    field: model.Field,
    value: Optional[Expr] = None,
    aggregate: Optional[Expr] = None,
    size: Optional[Expr] = None,
) -> Expr:
    package = prefix * message.identifier
    if value is None:
        value = Number(0)
    if aggregate is None:
        aggregate = EMPTY_ARRAY
    if size is None:
        size = Call(
            package * "Field_Size",
            [Variable("Ctx"), Variable(package * field.affixed_name)],
        )
    return Call(
        package * "Field_Condition",
        [
            Variable("Ctx"),
            Variable(package * field.affixed_name),
            *([value] if has_value_dependent_condition(message) else []),
            *([aggregate] if has_aggregate_dependent_condition(message) else []),
            *([size] if has_size_dependent_condition(message, field) else []),
        ],
    )


def to_base_integer(prefix: str, type_package: ID) -> ID:
    if type_package == BUILTINS_PACKAGE:
        return ID("To_Base_Integer")
    return prefix * type_package * "To_Base_Integer"


def ada_type_identifier(type_identifier: ID) -> ID:
    if model.is_builtin_type(type_identifier):
        return type_identifier.name

    return type_identifier


def prefixed_type_identifier(type_identifier: ID, prefix: str) -> ID:
    if model.is_builtin_type(type_identifier):
        return type_identifier.name

    return prefix * type_identifier


def enum_name(enum_type: model.Enumeration) -> ID:
    return ID(enum_type.name) + "_Enum"


def full_enum_name(enum_type: model.Enumeration) -> ID:
    return enum_type.identifier + "_Enum"


def sequence_name(message: model.Message, field: model.Field) -> ID:
    return message.types[field].identifier


def contains_function_name(refinement_package: ID, pdu: ID, sdu: ID, field: ID) -> str:
    sdu_name = sdu.name if sdu.parent == refinement_package else sdu
    pdu_name = pdu.name if pdu.parent == refinement_package else pdu
    return f"{sdu_name.flat}_In_{pdu_name.flat}_{field}"


def has_value_dependent_condition(
    message: model.Message,
    field: Optional[model.Field] = None,
) -> bool:
    links = message.outgoing(field) if field else message.structure
    fields = [field] if field else message.fields
    return any(
        r
        for l in links
        for r in l.condition.findall(lambda x: isinstance(x, expr.Relation))
        if isinstance(r, expr.Relation)
        and not r.findall(lambda x: isinstance(x, expr.Aggregate))
        and r.findall(
            lambda x: isinstance(x, expr.Variable)
            and any(x.identifier == f.identifier for f in fields),
        )
    )


def has_aggregate_dependent_condition(
    message: model.Message,
    field: Optional[model.Field] = None,
) -> bool:
    links = message.outgoing(field) if field else message.structure
    fields = [field] if field else message.fields
    return any(
        r
        for l in links
        for r in l.condition.findall(lambda x: isinstance(x, (expr.Equal, expr.NotEqual)))
        if isinstance(r, (expr.Equal, expr.NotEqual))
        and r.findall(lambda x: isinstance(x, expr.Aggregate))
        and any(
            r.left == expr.Variable(f.identifier) or r.right == expr.Variable(f.identifier)
            for f in fields
        )
    )


def has_size_dependent_condition(
    message: model.Message,
    field: Optional[model.Field] = None,
) -> bool:
    field_sizes = {expr.Size(f.name) for f in message.fields}
    links = message.outgoing(field) if field else message.structure
    return any(
        size in field_sizes
        for link in links
        for size in link.condition.findall(lambda x: isinstance(x, expr.Size))
    )


def create_sequence_instantiation(
    sequence_type: model.Sequence,
    prefix: str = "",
    flat: bool = False,
) -> tuple[list[ContextItem], GenericPackageInstantiation]:
    element_type = sequence_type.element_type
    element_type_package = element_type.package.name

    sequence_context: list[ContextItem] = []
    sequence_package: GenericPackageInstantiation
    if isinstance(element_type, model.Message):
        element_type_identifier = ID(
            element_type.identifier.flat if flat else prefix * element_type.identifier,
        )
        sequence_context = [
            WithClause(prefix * const.MESSAGE_SEQUENCE_PACKAGE),
            *([] if flat else [WithClause(element_type_identifier)]),
        ]
        sequence_package = GenericPackageInstantiation(
            ID(sequence_type.identifier.flat if flat else prefix * sequence_type.identifier),
            prefix * const.MESSAGE_SEQUENCE_PACKAGE,
            [
                element_type_identifier * "Context",
                element_type_identifier * "Initialize",
                element_type_identifier * "Take_Buffer",
                element_type_identifier * "Copy",
                element_type_identifier * "Has_Buffer",
                element_type_identifier * "Size",
                element_type_identifier * "Message_Last",
                element_type_identifier * "Initialized",
                element_type_identifier * "Well_Formed_Message",
            ],
        )
    elif isinstance(element_type, model.Scalar):
        element_type_identifier = prefix * element_type.identifier
        sequence_context = [
            WithClause(prefix * const.SCALAR_SEQUENCE_PACKAGE),
            *(
                [WithClause(prefix * element_type_package)]
                if element_type_package != sequence_type.package
                else []
            ),
        ]
        sequence_package = GenericPackageInstantiation(
            ID(sequence_type.identifier.flat if flat else prefix * sequence_type.identifier),
            prefix * const.SCALAR_SEQUENCE_PACKAGE,
            [
                element_type_identifier,
                str(element_type.size),
                prefix * element_type_package * f"Valid_{element_type.name}",
                prefix * element_type_package * "To_Actual",
                prefix * element_type_package * "To_Base_Integer",
            ],
        )
    else:
        assert False, 'unexpected element type "{type(element_type)}"'

    return (sequence_context, sequence_package)


def unchanged_cursor_before_or_invalid(
    limit: Expr,
    loop_entry: bool,
    or_invalid: bool = True,
    including_limit: bool = False,
) -> Expr:
    return ForAllIn(
        "F",
        Variable("Field"),
        IfExpr(
            [
                (
                    (
                        LessEqual(Variable("F"), limit)
                        if including_limit
                        else Less(Variable("F"), limit)
                    ),
                    Equal(
                        Indexed(
                            Variable("Ctx.Cursors"),
                            Variable("F"),
                        ),
                        Indexed(
                            (
                                LoopEntry(Variable("Ctx.Cursors"))
                                if loop_entry
                                else Old(Variable("Ctx.Cursors"))
                            ),
                            Variable("F"),
                        ),
                    ),
                ),
            ],
            *(
                [
                    Call(
                        "Invalid",
                        [
                            Variable("Ctx"),
                            Variable("F"),
                        ],
                    ),
                ]
                if or_invalid
                else []
            ),
        ),
    )


def conditional_field_size(
    field: model.Field,
    message: model.Message,
    prefix: str,
) -> Expr:
    def substituted(expression: expr.Expr) -> Expr:
        return (
            expression.substituted(
                substitution(
                    message,
                    prefix,
                    target_type=const.TYPES_BIT_LENGTH,
                    embedded=True,
                ),
            )
            .simplified()
            .ada_expr()
        )

    field_type = message.field_types[field]

    if isinstance(field_type, model.Scalar):
        return field_type.size.ada_expr()

    links = message.incoming(field)

    if len(links) == 1:
        return substituted(links[0].size)

    return If(
        [
            (
                AndThen(
                    Call(
                        "Well_Formed",
                        [Indexed(Variable("Cursors"), Variable(l.source.affixed_name))],
                    ),
                    *([substituted(l.condition)] if l.condition != expr.TRUE else []),
                ),
                substituted(l.size),
            )
            for l in links
        ],
        const.UNREACHABLE,
    )


def message_parameters(message: model.Message) -> list[Parameter]:
    return [
        Parameter([p.name], ada_type_identifier(t.identifier))
        for p, t in message.parameter_types.items()
    ]


def initialize_conditions(message: model.Message) -> list[Expr]:
    return [
        *[
            Equal(
                Variable("Ctx" * ID(p.name)),
                Variable(p.name),
            )
            for p, t in message.parameter_types.items()
        ],
        Call("Initialized", [Variable("Ctx")]),
    ]


def context_cursors_initialization(message: model.Message) -> Expr:
    return NamedAggregate(
        (
            message.fields[0].affixed_name,
            NamedAggregate(
                ("State", Variable("S_Invalid")),
                ("others", Variable("<>")),
            ),
        ),
        ("others", Variable("<>")),
    )


def byte_aligned_field(prefix: str, message: model.Message, field: model.Field) -> Expr:
    return Equal(
        Rem(
            Call(
                prefix * message.identifier * "Field_First",
                [
                    Variable("Ctx"),
                    Variable(prefix * message.identifier * field.affixed_name),
                ],
            ),
            Size(const.TYPES_BYTE),
        ),
        Number(1),
    )


def wrap_warning(expr_list: list[Declaration], warnings: list[str]) -> list[Declaration]:
    return [
        *[
            Pragma(
                "Warnings",
                [Variable("Off"), String(s)],
            )
            for s in warnings
        ],
        *expr_list,
        *[
            Pragma(
                "Warnings",
                [Variable("On"), String(s)],
            )
            for s in warnings
        ],
    ]


def comment_box(lines: list[str], width: int = 78) -> str:
    result = f"{'-' * width}\n"
    for line in lines:
        if len(line) == 0:
            result += f"--{' ' * (width - 5)} --\n"

        elif len(line) <= width - 6:
            space = width - 6 - len(line)
            left_space = " " * math.floor(space / 2)
            right_space = " " * math.ceil(space / 2)
            result += f"-- {left_space}{line}{right_space} --\n"

        else:
            for l in textwrap.wrap(line, width=width - 6):
                space = width - 6 - len(l)
                result += f"-- {l}{' ' * space} --\n"

    result += f"{'-' * width}\n"
    return result



from __future__ import annotations

import argparse
import json
import logging
import os
import re
import shutil
import subprocess
import sys
import traceback
from collections.abc import Sequence
from enum import Enum
from importlib import metadata
from multiprocessing import cpu_count
from pathlib import Path
from typing import Optional, Union

import importlib_resources
from importlib_resources.abc import Traversable

from rflx import __version__
from rflx.common import assert_never
from rflx.converter import iana
from rflx.error import (
    ERROR_CONFIG,
    FatalError,
    RecordFluxError,
    Severity,
    Subsystem,
    fail,
    fatal_fail,
)
from rflx.generator import Debug, Generator, optimizer
from rflx.graph import create_message_graph, create_session_graph, write_graph
from rflx.identifier import ID
from rflx.integration import Integration
from rflx.ls.server import server
from rflx.model import AlwaysVerify, Cache, Message, Model, NeverVerify, Session
from rflx.pyrflx import PyRFLXError
from rflx.specification import Parser
from rflx.validator import ValidationError, Validator

logging.basicConfig(level=logging.INFO, format="%(message)s")

DEFAULT_PREFIX = "RFLX"


class IDE(Enum):
    GNATSTUDIO = "gnatstudio"
    VSCODE = "vscode"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def ide(string: str) -> IDE:
        try:
            return IDE[string.upper()]
        except KeyError:
            raise ValueError from None


class UniqueStore(argparse.Action):
    """
    Action which allows at most one occurrence of the given option-value pair.

    The argparse module doesn't check for duplicate optional arguments having action 'store' (the
    default). However, such usage doesn't make sense (the value of the last matching argument
    remains, others are overwritten and ignored). This custom action class mimics the 'store' action
    in every other respect, but additionally prohibits such usage.

    Note: In argparse the 'store' action logic is implemented in argparse._StoreAction. However,
    since that is a protected class we can't directly inherit from it.
    """

    def __init__(  # type: ignore[no-untyped-def] # noqa: PLR0913
        self,
        option_strings,
        dest,
        nargs=None,
        const=None,
        default=None,
        type=None,  # noqa: A002
        choices=None,
        required=False,
        help=None,  # noqa: A002
        metavar=None,
    ):
        # Keep track whether the option has been set already
        self.is_set = False
        # Perform the 'store' action-related checks.
        # Note: This should be equivalent to the logic in argparse._StoreAction__init__(..).
        if nargs == 0:
            raise ValueError(
                "nargs for store actions must be != 0; if you "
                "have nothing to store, actions such as store "
                "true or store const may be more appropriate",
            )
        if const is not None and nargs != argparse.OPTIONAL:
            raise ValueError("nargs must be %r to supply const" % argparse.OPTIONAL)

        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(  # type: ignore [no-untyped-def]
        self,
        parser,
        namespace,
        values,
        option_string=None,
    ):
        if self.is_set:
            parser.error(f"{option_string} appears several times")
        else:
            self.is_set = True
            # Perform the 'store' action.
            # Note: This should be equivalent to the logic in argparse._StoreAction__call__(..).
            setattr(namespace, self.dest, values)


def run() -> Union[int, str]:
    return main(sys.argv)


def main(  # noqa: PLR0915
    argv: Sequence[str],
) -> Union[int, str]:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="disable logging to standard output",
    )
    parser.add_argument("--version", action="store_true")
    parser.add_argument(
        "--no-caching",
        action="store_true",
        help=("ignore verification cache"),
    )
    parser.add_argument(
        "--no-verification",
        action="store_true",
        help=("skip time-consuming verification of model"),
    )
    parser.add_argument(
        "--max-errors",
        action=UniqueStore,
        type=int,
        default=0,
        metavar=("NUM"),
        help="exit after at most NUM errors",
    )
    parser.add_argument(
        "--workers",
        action=UniqueStore,
        type=int,
        default=cpu_count(),
        metavar=("NUM"),
        help="parallelize proofs among NUM workers (default: %(default)d)",
    )
    parser.add_argument(
        "--unsafe",
        action="store_true",
        help="allow unsafe options (WARNING: may lead to erronous behavior)",
    )

    subparsers = parser.add_subparsers(dest="subcommand")

    parser_check = subparsers.add_parser("check", help="check specification")
    parser_check.add_argument(
        "files",
        metavar="SPECIFICATION_FILE",
        type=Path,
        nargs="+",
        help="specification file",
    )
    parser_check.set_defaults(func=check)

    parser_generate = subparsers.add_parser("generate", help="generate code")
    parser_generate.add_argument(
        "-p",
        "--prefix",
        action=UniqueStore,
        type=str,
        default=DEFAULT_PREFIX,
        help="add prefix to generated packages (default: %(default)s)",
    )
    parser_generate.add_argument(
        "-n",
        "--no-library",
        help="omit generating library files",
        action="store_true",
    )
    parser_generate.add_argument(
        "-d",
        action=UniqueStore,
        dest="output_directory",
        type=Path,
        default=".",
        help="output directory",
    )
    parser_generate.add_argument(
        "--debug",
        action=UniqueStore,
        default=None,
        choices=["built-in", "external"],
        help="enable adding of debug output to generated code",
    )
    parser_generate.add_argument(
        "--ignore-unsupported-checksum",
        help="ignore checksum aspects during code generation",
        action="store_true",
    )
    parser_generate.add_argument(
        "--integration-files-dir",
        action=UniqueStore,
        help="directory for the .rfi files",
        type=Path,
    )
    parser_generate.add_argument(
        "--reproducible",
        action="store_true",
        help="ensure reproducible output",
    )
    parser_generate.add_argument(
        "--optimize",
        action="store_true",
        help="optimize generated state machine code (requires GNATprove)",
    )
    parser_generate.add_argument(
        "--timeout",
        type=int,
        default=1,
        help="prover timeout in seconds for code optimization (default: %(default)s)",
    )
    parser_generate.add_argument(
        "files",
        metavar="SPECIFICATION_FILE",
        type=Path,
        nargs="*",
        help="specification file",
    )
    parser_generate.set_defaults(func=generate)

    parser_optimize = subparsers.add_parser(
        "optimize",
        help="optimize generated state machine code",
    )
    parser_optimize.add_argument(
        "--timeout",
        type=int,
        default=1,
        help="prover timeout in seconds (default: %(default)s)",
    )
    parser_optimize.add_argument(
        "directory",
        metavar="DIRECTORY",
        type=Path,
        help="directory containing the generated code",
    )
    parser_optimize.set_defaults(func=optimize)
    parser_graph = subparsers.add_parser("graph", help="generate graphs")
    parser_graph.add_argument(
        "-f",
        "--format",
        action=UniqueStore,
        type=str,
        default="svg",
        choices=["dot", "jpg", "pdf", "png", "raw", "svg"],
        help="output format (default: %(default)s)",
    )
    parser_graph.add_argument(
        "-i",
        "--ignore",
        type=str,
        metavar="REGEX",
        action="append",
        help="ignore states with names matching regular expression",
    )
    parser_graph.add_argument(
        "files",
        metavar="SPECIFICATION_FILE",
        type=Path,
        nargs="+",
        help="specification file",
    )
    parser_graph.add_argument(
        "-d",
        action=UniqueStore,
        dest="output_directory",
        type=Path,
        default=".",
        help="output directory",
    )
    parser_graph.set_defaults(func=graph)

    parser_validate = subparsers.add_parser(
        "validate",
        help="validate specification against a set of known valid or invalid messages",
    )
    parser_validate.add_argument(
        "--split-disjunctions",
        action="store_true",
        help=("split disjunctions before model validation (may have severe performance impact)"),
    )
    parser_validate.add_argument(
        "specification",
        metavar="SPECIFICATION_FILE",
        type=Path,
        help="specification file",
    )
    parser_validate.add_argument(
        "message_identifier",
        metavar="MESSAGE_IDENTIFIER",
        type=str,
        help="identifier of the top-level message (e.g., Package::Message)",
    )
    parser_validate.add_argument(
        "-v",
        action="append",
        dest="valid_sample_path",
        type=Path,
        help="known valid sample file or directory",
        default=None,
    )
    parser_validate.add_argument(
        "-i",
        action="append",
        dest="invalid_sample_path",
        type=Path,
        help="known invalid sample file or directory",
        default=None,
    )
    parser_validate.add_argument(
        "-c",
        action=UniqueStore,
        dest="checksum_module",
        type=str,
        help="name of the module containing the checksum functions",
        default=None,
    )
    parser_validate.add_argument(
        "-o",
        action=UniqueStore,
        dest="output_file",
        type=Path,
        help="path to output file for validation report in JSON format (file must not exist)",
        default=None,
    )
    parser_validate.add_argument(
        "--abort-on-error",
        action="store_true",
        help=(
            "abort with exitcode 1 if a message is classified as a false positive or false negative"
        ),
    )
    parser_validate.add_argument(
        "--coverage",
        action="store_true",
        help=(
            "enable coverage calculation and print the combined link coverage"
            " of all provided messages"
        ),
    )
    parser_validate.add_argument(
        "--target-coverage",
        action=UniqueStore,
        metavar="PERCENTAGE",
        type=float,
        default=0,
        help="abort with exitcode 1 if the coverage threshold is not reached",
    )
    parser_validate.set_defaults(func=validate)

    parser_install = subparsers.add_parser("install", help="set up RecordFlux IDE integration")
    parser_install.add_argument(
        "ide",
        type=IDE.ide,
        choices=[
            ide
            for ide in IDE
            if ide != IDE.VSCODE or vscode_extension().is_file() or os.environ.get("RFLX_TESTING")
        ],
    )
    parser_install.add_argument(
        "--gnat-studio-dir",
        action=UniqueStore,
        dest="gnat_studio_dir",
        type=Path,
        help="path to the GNAT Studio settings directory (default: $HOME/.gnatstudio)",
        default=Path.home() / ".gnatstudio",
    )
    parser_install.set_defaults(func=install)

    parser_convert = subparsers.add_parser(
        "convert",
        help="convert foreign specifications into RecordFlux specifications",
    )
    parser_convert.add_argument(
        "--reproducible",
        action="store_true",
        help="ensure reproducible output",
    )

    convert_subparsers = parser_convert.add_subparsers(dest="subcommand")

    parser_iana = convert_subparsers.add_parser(
        "iana",
        help="convert IANA registry into RecordFlux specifications",
    )

    parser_iana.add_argument(
        "-a",
        "--always-valid",
        help="add an Always_Valid aspect to each type that does not "
        "explicitly cover all possible values for the type's length",
        action="store_true",
    )
    parser_iana.add_argument(
        "-d",
        action=UniqueStore,
        dest="output_directory",
        type=Path,
        default=".",
        help="output directory",
    )
    parser_iana.add_argument("file", type=argparse.FileType("r"), nargs="?", default=sys.stdin)
    parser_iana.set_defaults(func=convert_iana)

    parser_run_ls = subparsers.add_parser("run_ls", help="run language server")
    parser_run_ls.set_defaults(func=run_language_server)

    args = parser.parse_args(argv[1:])

    if args.version:
        print(version())  # noqa: T201
        return 0

    if not args.subcommand:
        parser.print_usage()
        return 2

    if args.quiet:
        logging.disable(logging.CRITICAL)

    if not args.unsafe and args.no_verification:
        return 'cli: error: unsafe option "--no-verification" given without "--unsafe"'

    ERROR_CONFIG.fail_after_value = args.max_errors

    try:
        args.func(args)
    except RecordFluxError as e:
        return f"{e}"
    except Exception:  # noqa: BLE001
        if args.unsafe:
            return f"""
----------------------------------------------------------------------------
EXCEPTION IN UNSAFE MODE, PLEASE RERUN WITHOUT UNSAFE OPTIONS
----------------------------------------------------------------------------
{traceback.format_exc()}
----------------------------------------------------------------------------"""
        return f"""
------------------------------ RecordFlux Bug ------------------------------
{version()}

Optimized: {not __debug__}

Command: {' '.join(argv)}

{traceback.format_exc()}
----------------------------------------------------------------------------

A bug was detected. Please report this issue on GitHub:

https://github.com/AdaCore/RecordFlux/issues/new?labels=bug

Include the complete content of the bug box shown above and all input files
in the report."""

    return 0


def version() -> str:
    dependencies = [
        f"{r.name} {metadata.version(r.name)}"
        for r in (Requirement(r) for r in metadata.requires("RecordFlux") or [])
        if r.extra != "devel"
    ]
    return "\n".join(
        [
            f"RecordFlux {__version__}",
            *dependencies,
        ],
    )


class Requirement:
    def __init__(self, string: str) -> None:
        self.name: str
        self.extra: Optional[str]

        match = re.match(r'([^<=> (]{1,})[^;]*(?: *; extra == [\'"](.*)[\'"])?', string)

        if match:
            groups = match.groups()
            assert len(groups) == 2
            assert isinstance(groups[0], str)
            self.name = groups[0]
            self.extra = groups[1]
        else:
            fatal_fail(f'failed parsing requirement "{string}"', Subsystem.CLI)


def check(args: argparse.Namespace) -> None:
    parse(args.files, args.no_caching, args.no_verification, args.workers)


def generate(args: argparse.Namespace) -> None:
    # Eng/RecordFlux/Workarounds#28
    args.prefix = args.prefix if args.prefix != " " else ""

    if args.prefix and "" in args.prefix.split("."):
        fail(f'invalid prefix: "{args.prefix}"', Subsystem.CLI)

    if not args.output_directory.is_dir():
        fail(f'directory not found: "{args.output_directory}"', Subsystem.CLI)

    model, integration = parse(
        args.files,
        args.no_caching,
        args.no_verification,
        args.workers,
        args.integration_files_dir,
    )

    Generator(
        args.prefix,
        workers=args.workers,
        reproducible=args.reproducible,
        debug=(
            Debug.BUILTIN
            if args.debug == "built-in"
            else Debug.EXTERNAL if args.debug == "external" else Debug.NONE
        ),
        ignore_unsupported_checksum=args.ignore_unsupported_checksum,
    ).generate(
        model,
        integration,
        args.output_directory,
        library_files=not args.no_library,
        top_level_package=args.prefix == DEFAULT_PREFIX,
    )

    if args.optimize:
        optimizer.optimize(args.output_directory, args.workers, args.timeout)


def optimize(args: argparse.Namespace) -> None:
    if not args.directory.is_dir():
        fail(f'directory not found: "{args.directory}"', Subsystem.CLI)

    optimizer.optimize(args.directory, args.workers, args.timeout)


def parse(
    files: Sequence[Path],
    no_caching: bool,
    no_verification: bool,
    workers: int = 1,
    integration_files_dir: Optional[Path] = None,
) -> tuple[Model, Integration]:
    parser = Parser(
        cache(no_caching, no_verification),
        workers=workers,
        integration_files_dir=integration_files_dir,
    )
    error = RecordFluxError()
    present_files = []

    for f in files:
        if not f.is_file():
            error.extend([(f'file not found: "{f}"', Subsystem.CLI, Severity.ERROR, None)])
            continue

        present_files.append(Path(f))

    try:
        parser.parse(*present_files)
    except RecordFluxError as e:
        error.extend(e)

    try:
        model = parser.create_model()
    except RecordFluxError as e:
        error.extend(e)

    error.propagate()
    return model, parser.get_integration()


def graph(args: argparse.Namespace) -> None:
    if not args.output_directory.is_dir():
        fail(f'directory not found: "{args.output_directory}"', Subsystem.CLI)

    model, _ = parse(args.files, args.no_caching, args.no_verification)

    for d in model.declarations:
        filename = args.output_directory.joinpath(d.identifier.flat).with_suffix(f".{args.format}")
        if isinstance(d, Message):
            write_graph(create_message_graph(d), filename, fmt=args.format)
        if isinstance(d, Session):
            write_graph(create_session_graph(d, args.ignore), filename, fmt=args.format)

    locations: dict[str, dict[str, dict[str, dict[str, int]]]] = {
        str(package.location.source): {
            d.identifier.flat: {
                "start": {"line": d.location.start[0], "column": d.location.start[1]},
                "end": {"line": d.location.end[0], "column": d.location.end[1]},
            }
            for d in declarations
            if isinstance(d, (Message, Session)) and d.location and d.location.end
        }
        for package, declarations in model.packages.items()
        if package.location
    }

    filename = args.output_directory.joinpath("locations.json")
    with filename.open("w", encoding="utf-8") as f:
        json.dump(locations, f)


def validate(args: argparse.Namespace) -> None:
    try:
        identifier = ID(args.message_identifier)
    except FatalError as e:
        fail(f"invalid identifier: {e}", Subsystem.CLI)

    try:
        Validator(
            [args.specification],
            args.checksum_module,
            cache(args.no_caching, args.no_verification),
            split_disjunctions=args.split_disjunctions,
        ).validate(
            identifier,
            args.invalid_sample_path,
            args.valid_sample_path,
            args.output_file,
            args.abort_on_error,
            args.coverage,
            args.target_coverage,
        )
    except ValidationError as e:
        fail(str(e), Subsystem.VALIDATOR)
    except PyRFLXError as e:
        fatal_error = FatalError()
        fatal_error.extend(e)
        raise fatal_error from e


def install(args: argparse.Namespace) -> None:
    if args.ide is IDE.GNATSTUDIO:
        # TODO(eng/recordflux/RecordFlux#1359): Replace importlib_resources by importlib.resources
        gnatstudio_dir = importlib_resources.files("rflx") / "ide" / "gnatstudio"
        plugins_dir = args.gnat_studio_dir / "plug-ins"
        if not plugins_dir.exists():
            plugins_dir.mkdir(parents=True, exist_ok=True)
        print(f'Installing RecordFlux plugin into "{plugins_dir}"')  # noqa: T201
        shutil.copy(Path(gnatstudio_dir) / "recordflux.py", plugins_dir)

    elif args.ide is IDE.VSCODE:
        with importlib_resources.as_file(vscode_extension()) as extension:
            try:
                subprocess.run(["code", "--install-extension", extension, "--force"], check=True)
            except (FileNotFoundError, subprocess.CalledProcessError) as e:
                fail(f"installation of VS Code extension failed: {e}", Subsystem.CLI)

    else:
        assert_never(args.ide)  # pragma: no cover


def convert_iana(args: argparse.Namespace) -> None:
    xml_str = args.file.read()
    iana.convert(
        xml_str,
        args.file,
        args.always_valid,
        args.output_directory,
        args.reproducible,
    )


def run_language_server(args: argparse.Namespace) -> None:
    server.workers = args.workers
    server.start_io()


def vscode_extension() -> Traversable:
    # TODO(eng/recordflux/RecordFlux#1359): Replace importlib_resources by importlib.resources
    path = importlib_resources.files("rflx") / "ide" / "vscode" / "recordflux.vsix"
    assert isinstance(path, Traversable)
    return path


def cache(no_caching: bool, no_verification: bool) -> Cache:
    return NeverVerify() if no_verification else (AlwaysVerify() if no_caching else Cache())



from __future__ import annotations

from dataclasses import dataclass
from functools import singledispatchmethod
from typing import Optional, Union, cast

from rflx import const, lang
from rflx.identifier import ID

from .model import LSModel, Symbol, SymbolCategory


@dataclass
class Token:
    """
    Data structure produced by the LSLexer.

    Attributes
    ----------
        symbol: The symbol referenced by this token.
        lexeme: A string representation of the token in the file.
        line_number: An integer indicating the line at which the token is located in the file
                     starting from 1.
        character_offset: An integer indicating the character offset from the begining of the line
                          at which the first character of the lexeme is located starting from 1.

    """

    symbol: Optional[Symbol]
    lexeme: str
    line_number: int
    character_offset: int


@dataclass
class State:
    imports: set[ID]
    current_package: Optional[ID]
    declarations: list[ID]
    foreign_package: Optional[ID]
    current_session: Optional[ID]
    top_level: bool


class LSLexer:
    def __init__(self, model: LSModel):
        self._model = model
        self._tokens: list[Token] = []

    @property
    def tokens(self) -> list[Token]:
        """List of tokens resulting from the previous LSLexer.tokenize calls."""
        return self._tokens

    def search_token(self, line_number: int, character_offset: int) -> Optional[Token]:
        """Return the token at the given location if it exists, otherwise return None."""

        if len(self._tokens) == 0:
            return None

        lower_bound = 0
        upper_bound = len(self._tokens)

        # binary search over the lines
        while upper_bound - lower_bound > 1:
            middle = (lower_bound + upper_bound) // 2
            token = self._tokens[middle]

            if line_number < token.line_number:
                upper_bound = middle
            elif line_number > token.line_number:
                lower_bound = middle
            else:
                lower_bound = middle
                break

        # early exit if there are no tokens on requested line_number
        if self._tokens[lower_bound].line_number != line_number:
            return None

        # get to the first token of the line because the binary search
        # is likely to provide one in the middle
        while lower_bound > 0 and self._tokens[lower_bound - 1].line_number == line_number:
            lower_bound -= 1

        # linear search to find the token at the desired character offset if it exists
        while (
            lower_bound < len(self._tokens) and self._tokens[lower_bound].line_number == line_number
        ):
            if self._tokens[lower_bound].character_offset > character_offset:
                break
            if (
                self._tokens[lower_bound].character_offset + len(self._tokens[lower_bound].lexeme)
                >= character_offset
            ):
                return self._tokens[lower_bound]
            lower_bound += 1

        return None

    def tokenize(self, source: str, path: str = "<stdin>") -> None:
        """Convert a string into a list of tokens that can be accessed via the tokens property."""

        unit = lang.AnalysisContext().get_from_buffer(
            path,
            source,
            rule=lang.GrammarRule.main_rule_rule,
        )

        state = State(set(), None, [], None, None, top_level=False)
        self._process_ast_node(unit.root, state)

    @singledispatchmethod
    def _process_ast_node(self, node: Optional[lang.RFLXNode], state: State) -> None:
        if node is None:
            return

        self._process_children(node, state)

    def _process_children(self, node: lang.RFLXNode, state: State) -> None:
        for child in node.children:
            self._process_ast_node(child, state)

    def _identify_symbol(self, lexeme: str, state: State) -> Optional[Symbol]:
        symbols: list[Symbol] = self._model.get_symbols(lexeme)

        if state.foreign_package is not None and state.foreign_package != ID(lexeme):
            for symbol in symbols:
                if symbol.identifier == state.foreign_package * ID(lexeme):  # pragma: no branch
                    return symbol
            return None

        if not state.top_level:
            for symbol in symbols:
                if symbol.parent in state.declarations:
                    return symbol

        for symbol in filter(lambda s: s.parent is None, symbols):
            if (
                (
                    symbol.category == SymbolCategory.PACKAGE
                    and (
                        symbol.identifier in state.imports
                        or symbol.identifier == state.current_package
                    )
                )
                or (
                    state.current_package is not None
                    and symbol.identifier == state.current_package * ID(lexeme)
                )
                or symbol.identifier == const.BUILTINS_PACKAGE * ID(lexeme)
                or symbol.identifier == const.INTERNAL_PACKAGE * ID(lexeme)
            ):
                return symbol

        return None

    @_process_ast_node.register
    def _(self, node: lang.ID, state: State) -> None:
        if node.f_package is not None:
            state.foreign_package = ID(node.f_package.text)
        self._process_children(node, state)
        state.foreign_package = None

    @_process_ast_node.register
    def _(self, node: lang.UnqualifiedID, state: State) -> None:
        symbol = self._identify_symbol(node.text, state)

        self._tokens.append(
            Token(
                symbol,
                node.text,
                node.sloc_range.start.line - 1,
                node.sloc_range.start.column - 1,
            ),
        )

    @_process_ast_node.register
    def _(self, node: lang.VariableDecl, state: State) -> None:
        self._process_ast_node(node.f_identifier, state)
        state.top_level = True
        self._process_ast_node(node.f_type_identifier, state)
        state.top_level = False

    @_process_ast_node.register
    def _(self, node: lang.Parameter, state: State) -> None:
        self._process_ast_node(node.f_identifier, state)
        state.top_level = True
        self._process_ast_node(node.f_type_identifier, state)
        state.top_level = False

    @_process_ast_node.register
    def _(self, node: lang.ContextItem, state: State) -> None:
        state.imports.add(ID(node.f_item.text))
        self._process_children(node, state)

    @_process_ast_node.register
    def _(self, node: lang.PackageNode, state: State) -> None:
        state.current_package = ID(node.f_identifier.text)
        self._process_children(node, state)
        state.current_package = None

    # Python 3.11 directly supports single dispatch with typing.Union
    @_process_ast_node.register(lang.TypeDecl)
    @_process_ast_node.register(lang.SessionDecl)
    def _(self, node: Union[lang.TypeDecl, lang.SessionDecl], state: State) -> None:
        partial_identifier = ID(node.f_identifier.text)
        identifier = (
            state.current_package * partial_identifier
            if state.current_package is not None
            else partial_identifier
        )

        if node.is_a(lang.SessionDecl):
            state.current_session = identifier

        state.declarations.append(identifier)
        self._process_children(node, state)
        state.declarations.pop()

        if node.is_a(lang.SessionDecl):
            state.current_session = None

    @_process_ast_node.register
    def _(self, node: lang.MessageField, state: State) -> None:
        self._process_ast_node(node.f_identifier, state)
        state.top_level = True
        self._process_ast_node(node.f_type_identifier, state)
        state.top_level = False
        self._process_ast_node(node.f_aspects, state)
        self._process_ast_node(node.f_condition, state)
        self._process_ast_node(node.f_thens, state)
        self._process_ast_node(node.f_type_identifier, state)

    @_process_ast_node.register
    def _(self, node: lang.RefinementDecl, state: State) -> None:
        name = node.f_pdu.f_name
        # TODO(eng/recordflux/RecordFlux#1371): Invalid type annotation for optional field
        if cast(Optional[lang.UnqualifiedID], node.f_pdu.f_package) is None:
            message_identifier = (
                state.current_package * ID(name.text)
                if state.current_package is not None
                else ID(name.text)
            )
        else:
            message_identifier = ID(node.f_pdu.f_package.text) * ID(name.text)
        state.declarations.append(message_identifier)
        self._process_children(node, state)
        state.declarations.pop()

    @_process_ast_node.register
    def _(self, node: lang.FormalFunctionDecl, state: State) -> None:
        assert state.current_session is not None
        state.declarations.append(state.current_session * ID(node.f_identifier.text))
        self._process_children(node, state)
        state.declarations.pop()

    @_process_ast_node.register
    def _(self, node: lang.State, state: State) -> None:
        assert state.current_session is not None
        state.declarations.append(state.current_session * ID(node.f_identifier.text))
        self._process_children(node, state)
        state.declarations.pop()



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from itertools import zip_longest
from typing import Optional

from rflx import ir, typing_ as rty
from rflx.ada import (
    Add,
    And,
    AndThen,
    Assignment,
    Call,
    Component,
    ContextItem,
    DynamicPredicate,
    Equal,
    ExpressionFunctionDeclaration,
    First,
    FunctionSpecification,
    InOutParameter,
    Last,
    NamedAggregate,
    NotEqual,
    NullStatement,
    Number,
    OrElse,
    OutParameter,
    Parameter,
    Postcondition,
    ProcedureSpecification,
    RecordType,
    Slice,
    SparkMode,
    SubprogramBody,
    SubprogramDeclaration,
    Subtype,
    UnitPart,
    UnrestrictedAccess,
    UseTypeClause,
    Variable,
    WithClause,
)
from rflx.error import Location
from rflx.identifier import ID
from rflx.integration import Integration

from . import const


@dataclass
class SlotInfo:
    size: int
    locations: Sequence[Location]


@dataclass
class NumberedSlotInfo(SlotInfo):
    slot_id: int
    global_: bool


class AllocatorGenerator:
    def __init__(self, session: ir.Session, integration: Integration, prefix: str = "") -> None:
        self._session = session
        self._prefix = prefix
        self._declaration_context: list[ContextItem] = []
        self._body_context: list[ContextItem] = []
        self._allocation_slots: dict[Location, int] = {}
        self._unit_part = UnitPart()
        self._integration = integration

        global_slots: list[SlotInfo] = self._allocate_global_slots()
        local_slots: list[SlotInfo] = self._allocate_local_slots()
        numbered_slots: list[NumberedSlotInfo] = []
        count = 1

        for slot in global_slots:
            numbered_slots.append(NumberedSlotInfo(slot.size, slot.locations, count, global_=True))
            count += 1
        for slot in local_slots:
            numbered_slots.append(NumberedSlotInfo(slot.size, slot.locations, count, global_=False))
            count += 1
        for slot in numbered_slots:
            for location in slot.locations:
                self._allocation_slots[location] = slot.slot_id

        self._create(numbered_slots)

        self._numbered_slots = numbered_slots

    @property
    def unit_identifier(self) -> ID:
        return self._session.identifier.parent * f"{self._session.identifier.name}_Allocator"

    @property
    def declaration_context(self) -> list[ContextItem]:
        return self._declaration_context

    @property
    def body_context(self) -> list[ContextItem]:
        return self._body_context

    @property
    def unit_part(self) -> UnitPart:
        return self._unit_part

    @property
    def required(self) -> bool:
        return bool(self._allocation_slots)

    def get_global_slot_ptrs(self) -> list[ID]:
        return [self._slot_name(s.slot_id) for s in self._numbered_slots if s.global_]

    def get_local_slot_ptrs(self) -> list[ID]:
        return [self._slot_name(s.slot_id) for s in self._numbered_slots if not s.global_]

    def get_slot_ptr(self, location: Optional[Location]) -> ID:
        assert location is not None
        slot_id: int = self._allocation_slots[location]
        return self._slot_name(slot_id)

    def get_size(self, variable: Optional[ID] = None, state: Optional[ID] = None) -> int:
        return self._integration.get_size(self._session.identifier, variable, state)

    @staticmethod
    def _slot_name(slot_id: int) -> ID:
        return ID(f"Slot_Ptr_{slot_id}")

    @staticmethod
    def _ptr_type(size: int) -> ID:
        return ID(f"Slot_Ptr_Type_{size}")

    def _create(self, slots: Sequence[NumberedSlotInfo]) -> None:
        self._unit_part += self._create_memory(slots)
        if slots:
            self._unit_part += self._create_ptr_subtypes(slots)
        self._unit_part += self._create_slots(slots)
        self._unit_part += self._create_init_pred(slots)
        self._unit_part += self._create_uninitialized_pred(slots)
        self._unit_part += self._create_init_proc(slots)
        self._unit_part += self._create_finalize_proc(slots)
        self._unit_part += self._create_global_allocated_pred(slots)

    @staticmethod
    def _create_memory(slots: Sequence[NumberedSlotInfo]) -> UnitPart:
        return UnitPart(
            [
                RecordType(
                    "Memory",
                    [
                        Component(
                            f"Slot_{slot.slot_id}",
                            Slice(
                                Variable(const.TYPES_BYTES),
                                First(const.TYPES_INDEX),
                                Add(First(const.TYPES_INDEX), Number(slot.size - 1)),
                            ),
                            NamedAggregate(("others", Number(0))),
                            aliased=True,
                        )
                        for slot in slots
                    ],
                ),
            ],
        )

    def _create_ptr_subtypes(self, slots: Sequence[NumberedSlotInfo]) -> UnitPart:
        unit = UnitPart(
            specification=[
                Subtype(
                    self._ptr_type(size),
                    const.TYPES_BYTES_PTR,
                    aspects=[
                        DynamicPredicate(
                            OrElse(
                                Equal(Variable(self._ptr_type(size)), Variable("null")),
                                AndThen(
                                    Equal(First(self._ptr_type(size)), First(const.TYPES_INDEX)),
                                    Equal(
                                        Last(self._ptr_type(size)),
                                        Add(First(const.TYPES_INDEX), Number(size - 1)),
                                    ),
                                ),
                            ),
                        ),
                    ],
                )
                for size in sorted({slot.size for slot in slots})
            ],
        )
        self._declaration_context.append(WithClause(self._prefix * const.TYPES_PACKAGE))
        self._declaration_context.append(UseTypeClause(self._prefix * const.TYPES_INDEX))
        self._declaration_context.append(UseTypeClause(self._prefix * const.TYPES_BYTES_PTR))
        return unit

    def _create_slots(self, slots: Sequence[NumberedSlotInfo]) -> UnitPart:
        self._declaration_context.append(WithClause(self._prefix * const.TYPES_PACKAGE))

        return UnitPart(
            [
                RecordType(
                    "Slots",
                    [
                        Component(self._slot_name(slot.slot_id), self._ptr_type(slot.size))
                        for slot in slots
                    ],
                ),
            ],
        )

    def _create_init_pred(self, slots: Sequence[NumberedSlotInfo]) -> UnitPart:
        return UnitPart(
            [
                ExpressionFunctionDeclaration(
                    FunctionSpecification("Initialized", "Boolean", [Parameter(["S"], "Slots")]),
                    And(
                        *[
                            NotEqual(
                                Variable("S" * self._slot_name(slot.slot_id)),
                                Variable("null"),
                            )
                            for slot in slots
                        ],
                    ),
                ),
            ],
        )

    def _create_uninitialized_pred(self, slots: Sequence[NumberedSlotInfo]) -> UnitPart:
        return UnitPart(
            [
                ExpressionFunctionDeclaration(
                    FunctionSpecification("Uninitialized", "Boolean", [Parameter(["S"], "Slots")]),
                    And(
                        *[
                            Equal(
                                Variable("S" * self._slot_name(slot.slot_id)),
                                Variable("null"),
                            )
                            for slot in slots
                        ],
                    ),
                ),
            ],
        )

    def _create_global_allocated_pred(self, slots: Sequence[NumberedSlotInfo]) -> UnitPart:
        return UnitPart(
            [
                ExpressionFunctionDeclaration(
                    FunctionSpecification(
                        "Global_Allocated",
                        "Boolean",
                        [Parameter(["S"], "Slots")],
                    ),
                    And(
                        *[
                            Equal(
                                Variable("S" * self._slot_name(slot.slot_id)),
                                Variable("null"),
                            )
                            for slot in slots
                            if slot.global_
                        ],
                        *[
                            NotEqual(
                                Variable("S" * self._slot_name(slot.slot_id)),
                                Variable("null"),
                            )
                            for slot in slots
                            if not slot.global_
                        ],
                    ),
                ),
            ],
        )

    def _create_init_proc(self, slots: Sequence[NumberedSlotInfo]) -> UnitPart:
        proc = ProcedureSpecification(
            "Initialize",
            [OutParameter(["S"], "Slots"), Parameter(["M"], "Memory")],
        )
        return UnitPart(
            [
                SubprogramDeclaration(proc, [Postcondition(Call("Initialized", [Variable("S")]))]),
            ],
            [
                SubprogramBody(
                    proc,
                    declarations=[],
                    statements=(
                        [
                            Assignment(
                                "S" * self._slot_name(slot.slot_id),
                                UnrestrictedAccess(Variable(ID(f"M.Slot_{slot.slot_id}"))),
                            )
                            for slot in slots
                        ]
                        if slots
                        else [NullStatement()]
                    ),
                    aspects=[SparkMode(off=True)],
                ),
            ],
        )

    def _create_finalize_proc(self, slots: Sequence[NumberedSlotInfo]) -> UnitPart:
        proc = ProcedureSpecification("Finalize", [InOutParameter(["S"], "Slots")])
        return UnitPart(
            [
                SubprogramDeclaration(
                    proc,
                    [Postcondition(Call("Uninitialized", [Variable("S")]))],
                ),
            ],
            [
                SubprogramBody(
                    proc,
                    declarations=[],
                    statements=(
                        [
                            Assignment(
                                "S" * self._slot_name(slot.slot_id),
                                Variable("null"),
                            )
                            for slot in slots
                        ]
                        if slots
                        else [NullStatement()]
                    ),
                    aspects=[SparkMode(off=True)],
                ),
            ],
        )

    @staticmethod
    def _needs_allocation(type_: rty.Type) -> bool:
        return isinstance(type_, (rty.Message, rty.Sequence))

    def _allocate_global_slots(self) -> list[SlotInfo]:
        slots = []
        for d in self._session.declarations:
            if self._needs_allocation(d.type_):
                assert d.location is not None
                slots.append(SlotInfo(self.get_size(d.identifier), [d.location]))
        return slots

    @staticmethod
    def _scope(state: ir.State, var_id: ID) -> Optional[ID]:
        """
        Return the scope of the variable var_id.

        Return the ID of the state if var_id was defined in that state, otherwise
        return None to indicate that var_id is a global variable of the session.
        """
        if any(var_id == d.identifier for d in state.declarations):
            return state.identifier.name
        return None

    def _allocate_local_slots(
        self,
    ) -> list[SlotInfo]:
        """
        Allocate slots for state variables and state actions.

        We collect the allocation requirements (pairs of locations and sizes) for each
        state in a list, then sort these lists in descending order of sizes.
        Then, we imagine a matrix where each line is one of these lists (filling
        up with zero sizes to make the lists the same length). For each column of
        that matrix, we create a SlotInfo object whose size is the maximal size for
        that column, and the locations are all locations in that column.
        """

        @dataclass
        class AllocationRequirement:
            location: Optional[Location]
            size: int

        def determine_allocation_requirements(
            statements: Sequence[ir.Stmt],
            state: ir.State,
        ) -> list[AllocationRequirement]:
            alloc_requirements = []

            for statement in statements:
                if isinstance(statement, ir.VarDecl) and self._needs_allocation(statement.type_):
                    alloc_requirements.append(
                        AllocationRequirement(
                            statement.location,
                            self.get_size(statement.identifier, state.identifier.name),
                        ),
                    )
                if (
                    isinstance(statement, ir.Assign)
                    and isinstance(statement.expression, ir.Comprehension)
                    and isinstance(statement.expression.sequence.type_, rty.Sequence)
                    and isinstance(statement.expression.sequence.type_.element, rty.Message)
                    and isinstance(statement.expression.sequence, (ir.Var, ir.FieldAccess))
                ):
                    if isinstance(statement.expression.sequence, ir.FieldAccess):
                        identifier = statement.expression.sequence.message
                    else:
                        assert isinstance(statement.expression.sequence, ir.Var)
                        identifier = statement.expression.sequence.identifier
                    alloc_requirements.append(
                        AllocationRequirement(
                            statement.location,
                            self.get_size(identifier, self._scope(state, identifier)),
                        ),
                    )
                if isinstance(statement, ir.Assign) and isinstance(statement.expression, ir.Head):
                    identifier = statement.expression.prefix
                    alloc_requirements.append(
                        AllocationRequirement(
                            statement.location,
                            self.get_size(identifier, self._scope(state, identifier)),
                        ),
                    )
                if isinstance(statement, ir.Assign) and isinstance(statement.expression, ir.Find):
                    if isinstance(statement.expression.sequence, ir.Var):
                        identifier = statement.expression.sequence.identifier
                    elif isinstance(statement.expression.sequence, ir.FieldAccess):
                        identifier = statement.expression.sequence.message
                    else:
                        assert False
                    alloc_requirements.append(
                        AllocationRequirement(
                            statement.location,
                            self.get_size(identifier, self._scope(state, identifier)),
                        ),
                    )
                if isinstance(statement, ir.Assign) and isinstance(
                    statement.expression,
                    (ir.Comprehension, ir.Find),
                ):
                    alloc_requirements.extend(
                        determine_allocation_requirements(
                            statement.expression.selector.stmts,
                            state,
                        ),
                    )
                    alloc_requirements.extend(
                        determine_allocation_requirements(
                            statement.expression.condition.stmts,
                            state,
                        ),
                    )

            return alloc_requirements

        alloc_requirements_per_state = [
            determine_allocation_requirements(state.actions, state)
            for state in self._session.states
        ]

        for state_requirements in alloc_requirements_per_state:
            state_requirements.sort(key=lambda x: x.size, reverse=True)

        slots = []
        for requirements_per_slot in zip_longest(
            *alloc_requirements_per_state,
            fillvalue=AllocationRequirement(None, 0),
        ):
            size = max(x.size for x in requirements_per_slot)
            locations = [x.location for x in requirements_per_slot if x.location is not None]
            slots.append(SlotInfo(size, locations))
        return slots



from __future__ import annotations

from collections.abc import Sequence
from typing import Union

from typing_extensions import Self

from rflx.pyrflx.error import PyRFLXError


class Bitstring:
    def __init__(self, bits: str = ""):
        if not self.valid_bitstring(bits):
            raise PyRFLXError("Bitstring does not consist of only 0 and 1")
        self._bits = bits

    def __add__(self, other: Bitstring) -> Bitstring:
        return Bitstring(self._bits + other._bits)

    def __iadd__(self, other: Bitstring) -> Self:
        self._bits += other._bits
        return self

    def __getitem__(self, key: Union[int, slice]) -> Bitstring:
        if isinstance(key, slice) and isinstance(key.stop, int) and len(self._bits) < key.stop:
            raise IndexError
        return Bitstring(self._bits[key])

    def __repr__(self) -> str:
        return f'Bitstring("{self._bits}")'

    def __str__(self) -> str:
        return self._bits

    def __int__(self) -> int:
        return int(self._bits, 2)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Bitstring):
            return NotImplemented
        return self._bits == other._bits

    def __bytes__(self) -> bytes:
        return b"".join(
            [
                int(self._bits[i : i + 8], 2).to_bytes(1, "big")
                for i in range(0, len(self._bits), 8)
            ],
        )

    def __len__(self) -> int:
        return len(self._bits)

    @staticmethod
    def swap_bitstring(bitstring: str) -> str:
        assert len(bitstring) > 8
        assert len(bitstring) % 8 == 0
        result = ""
        for i in range(len(bitstring), 0, -8):
            result += bitstring[i - 8 : i]
        assert len(bitstring) == len(result)
        return result

    def swap(self) -> Bitstring:
        return Bitstring(self.swap_bitstring(self._bits))

    @classmethod
    def from_bytes(cls, msg: bytes) -> Bitstring:
        return cls(format(int.from_bytes(msg, "big"), f"0{len(msg) * 8}b"))

    @staticmethod
    def valid_bitstring(bitstring: str) -> bool:
        return all(bit in ["0", "1"] for bit in bitstring)

    @staticmethod
    def join(iterable: Sequence[Bitstring]) -> Bitstring:
        joined_bitstring = Bitstring()
        for i in iterable:
            joined_bitstring += i

        return joined_bitstring
